#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the cscript base classes.
#
# Copyright (C) 2016-2019 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import ctools
from testing import test


# ================== #
# cscript_test class #
# ================== #
class cscript_test(ctools.cscript):
    """
    Test class for the cscript base class
    """
    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Return
        return

    # Check setup_observations() method
    def check_setup_observations(self, filename, events=True, cube=True):
        self.pars()['inobs'].value(filename)
        obs = gammalib.GObservations()
        self._setup_observations(obs, False, events, cube)
        return

    # Check setup_models() method
    def check_setup_models(self):
        obs = gammalib.GObservations()
        self._setup_models(obs)
        return

    # Check create_ebounds() method
    def check_create_ebounds(self, filename):
        if filename is not '':
            self.pars()['ebinfile'].value(filename)
        return self._create_ebounds()

    # Check create_energies() method
    def check_create_energies(self, filename):
        if filename is not '':
            self.pars()['ebinfile'].value(filename)
        return self._create_energies()

    # Check require_inobs() method
    def check_require_inobs(self, filename):
        self.pars()['inobs'].value(filename)
        self._require_inobs('check_require_inobs')
        return

    # Check require_inobs_nocube() method
    def check_require_inobs_nocube(self, filename):
        self.pars()['inobs'].value(filename)
        self._require_inobs_nocube('check_require_inobs_nocube')
        return

    # Check get_roi() method
    def check_get_roi(self):
        pnt = gammalib.GCTAPointing()
        return self._get_roi(pnt)

    # Check get_ebounds() method
    def check_get_ebounds(self):
        return self._get_ebounds()

    # Check get_gti() method
    def check_get_gti(self):
        ref = gammalib.GTimeReference()
        return self._get_gti(ref)

    # Check get_pointing() method
    def check_get_pointing(self):
        return self._get_pointing()

    # Check get_skydir() method
    def check_get_skydir(self):
        return self._get_skydir()

    # Check set_outfile_name() method
    def check_set_outfile_name(self, filename):
        return self._set_outfile_name(filename)

    # Check is_stacked() method
    def check_is_stacked(self):
        return self._is_stacked()

    # Check is_onoff() method
    def check_is_onoff(self):
        return self._is_onoff()

    # Check restore_edisp() method
    def check_restore_edisp(self):
        obs   = gammalib.GObservations()
        edisp = [True, True]
        self._restore_edisp(obs, edisp)
        return

    # Check get_mean_pointing() method
    def check_get_mean_pointing(self, obs):
        return self._get_mean_pointing(obs)

    # Check set_obs_response() method
    def check_set_obs_response(self, expcube, psfcube, bkgcube, edispcube, edisp):
        obs    = gammalib.GCTAObservation()
        counts = gammalib.GCTAEventCube()
        obs.events(counts)
        self.pars()['expcube'].value(expcube)
        self.pars()['psfcube'].value(psfcube)
        self.pars()['bkgcube'].value(bkgcube)
        self.pars()['edispcube'].value(edispcube)
        self.pars()['edisp'].value(edisp)
        self._set_obs_response(obs)
        return

    # Check get_current_rss() method
    def check_get_current_rss(self):
        return self._get_current_rss()

    # Check save_event_list() method
    def check_save_event_list(self, infile, evtname, gtiname, outfile):
        obs    = gammalib.GCTAObservation()
        events = gammalib.GCTAEventList()
        obs.events(events)
        obs.save(infile, True)
        self._save_event_list(obs, infile, evtname, gtiname, outfile)
        return

    # Check warn_too_few_energies() method
    def check_warn_too_few_energies(self, energies):
        return self._warn_too_few_energies(energies)

    # Check warn_xml_suffix() method
    def check_warn_xml_suffix(self, filename):
        return self._warn_xml_suffix(gammalib.GFilename(filename))


# ======================== #
# ctobservation_test class #
# ======================== #
class ctobservation_test(ctools.csobservation):
    """
    Test class for the ctobservation base class
    """
    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Return
        return

    # Check first_unbinned_observation() method
    def check_first_unbinned_observation(self):
        return self._first_unbinned_observation()

    # Check next_unbinned_observation() method
    def check_next_unbinned_observation(self):
        return self._next_unbinned_observation()

    # Check set_obs_statistic() method for unbinned observation
    def check_set_obs_statistic_unbinned(self, statistic):
        obs    = gammalib.GObservations()
        events = gammalib.GCTAEventList()
        cta    = gammalib.GCTAObservation()
        cta.events(events)
        obs.append(cta)
        self.obs(obs)
        self._set_obs_statistic(statistic)
        return gammalib.toupper(self.obs()[0].statistic())

    # Check set_obs_statistic() method for binned observation
    def check_set_obs_statistic_binned(self, statistic):
        obs  = gammalib.GObservations()
        cube = gammalib.GCTAEventCube()
        cta  = gammalib.GCTAObservation()
        cta.events(cube)
        obs.append(cta)
        self.obs(obs)
        self._set_obs_statistic(statistic)
        return gammalib.toupper(self.obs()[0].statistic())

    # Check set_obs_statistic() method for On/Off observation
    def check_set_obs_statistic_onoff(self, statistic):
        obs = gammalib.GObservations()
        cta = gammalib.GCTAOnOffObservation()
        obs.append(cta)
        self.obs(obs)
        self._set_obs_statistic(statistic)
        return gammalib.toupper(self.obs()[0].statistic())

    # Check set_obs_bounds() method
    def check_set_obs_bounds(self):
        obs    = gammalib.GObservations()
        events = gammalib.GCTAEventList()
        cta    = gammalib.GCTAObservation()
        cta.events(events)
        obs.append(cta)
        self.obs(obs)
        self._set_obs_bounds()
        return self.obs()

    # Check read_ogip_keywords() and write_ogip_keywords methods
    def check_ogip_keywords(self, hdu):
        self._read_ogip_keywords(hdu)
        self._write_ogip_keywords(hdu)
        self._read_ogip_keywords(hdu)
        return hdu


# ======================= #
# ctlikelihood_test class #
# ======================= #
class ctlikelihood_test(ctools.cslikelihood):
    """
    Test class for the ctlikelihood base class
    """
    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_cslikelihood(self.__class__.__name__, ctools.__version__, argv)

        # Return
        return

    # Check evaluate() method
    def check_evaluate(self, value):
        par = gammalib.GModelPar()
        par.value(2.0)
        par.range(1.0, 3.0)
        return self._evaluate(par, value)


# =========================== #
# Test class for base classes #
# =========================== #
class Test(test):
    """
    Test class for base classes
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Create parameter file for cscript_test application
        pars = gammalib.GApplicationPars()
        pars.append(gammalib.GApplicationPar('inobs','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('inmodel','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('ebinalg','s','h','FILE','','',''))
        pars.append(gammalib.GApplicationPar('ebinfile','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('ra','r','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('dec','r','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('rad','r','h','1.0','','',''))
        pars.append(gammalib.GApplicationPar('emin','r','h','1.0','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','10.0','','',''))
        pars.append(gammalib.GApplicationPar('tmin','t','h','2005-10-08T14:30:25','','',''))
        pars.append(gammalib.GApplicationPar('tmax','t','h','2005-10-08T14:58:26','','',''))
        pars.append(gammalib.GApplicationPar('expcube','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('psfcube','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('bkgcube','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('edispcube','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('edisp','b','h','yes','','',''))
        pars.append(gammalib.GApplicationPar('caldb','s','h','prod2','','',''))
        pars.append(gammalib.GApplicationPar('irf','s','h','South_0.5h','','',''))
        pars.append_standard()
        pars.save('cscript_test.par')

        # Create parameter file for ctobservation_test.par application
        pars = gammalib.GApplicationPars()
        pars.append_standard()
        pars.save('ctobservation_test.par')

        # Create parameter file for ctlikelihood_test application
        pars = gammalib.GApplicationPars()
        pars.append_standard()
        pars.save('ctlikelihood_test.par')

        # Create energy boundaries file
        ebounds = gammalib.GEbounds(1, gammalib.GEnergy(1.0,'TeV'),
                                       gammalib.GEnergy(10.0,'TeV'))
        ebounds.save('test_ebinfile_ebounds.fits[EBOUNDS]', True)
        ebounds.save('test_ebinfile_energybins.fits[ENERGYBINS]', True)
        ebounds.save('test_ebinfile_bins.fits[BINS]', True)

        # Create energy bin file
        energies = gammalib.GEnergies(2, gammalib.GEnergy(1.0,'TeV'),
                                         gammalib.GEnergy(10.0,'TeV'))
        energies.save('test_ebinfile_energies.fits[ENERGIES]', True)
        energies.save('test_ebinfile_engs.fits[ENGS]', True)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('cscript')

        # Append tests
        self.append(self._test_cscript, 'Test cscript base class')
        self.append(self._test_ctobservation, 'Test ctobservation base class')
        self.append(self._test_ctlikelihood, 'Test ctlikelihood base class')

        # Return
        return

    # Setup cscript class
    def _setup_cscript(self):
        """
        Setup cscript test class
        """
        # Get test class
        cls = cscript_test()

        # Recover User parameters
        pars = cls.pars()

        # Create a copy of the User parameters
        cpy_pars = pars.copy()

        # Clear User parameters
        pars.clear()

        # Return
        return (cls, pars, cpy_pars)

    # Setup ctobservation test class
    def _setup_ctobservation_test(self):
        """
        Setup cscript test class
        """
        # Get test class
        cls = ctobservation_test()

        # Recover User parameters
        pars = cls.pars()

        # Create a copy of the User parameters
        cpy_pars = pars.copy()

        # Clear User parameters
        pars.clear()

        # Return
        return (cls, pars, cpy_pars)

    # Test cscript base class
    def _test_cscript(self):
        """
        Test cscript base class
        """
        # Allocate empty
        empty = cscript_test()

        # Test read_ahead() method
        # TODO: implement

        # Test time_reference() method
        # TODO: implement

        # Test get_observations() method
        # TODO: implement

        # Test setup_observations() for invalid file name
        self.test_try('Check setup_observations() for "NONE"')
        try:
            empty.check_setup_observations('NONE')
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test setup_observations() for acceptable event list
        self.test_try('Check setup_observations() for acceptable event list')
        try:
            empty.check_setup_observations(self._events, events=True)
            self.test_try_success()
        except ValueError:
            self.test_try_failure('Exception thrown')

        # Test setup_observations() for unacceptable event list
        self.test_try('Check setup_observations() for unacceptable event list')
        try:
            empty.check_setup_observations(self._events, events=False)
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test setup_observations() for acceptable counts cube
        self.test_try('Check setup_observations() for acceptable counts cube')
        try:
            empty.check_setup_observations(self._cntcube, cube=True)
            self.test_try_success()
        except ValueError:
            self.test_try_failure('Exception thrown')

        # Test setup_observations() for unacceptable counts cube
        self.test_try('Check setup_observations() for unacceptable counts cube')
        try:
            empty.check_setup_observations(self._cntcube, cube=False)
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test setup_models() for invalid model definition file name
        self.test_try('Check setup_models() for invalid model definition file name')
        try:
            empty.check_setup_models()
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test create_ebounds() method for "EBOUNDS" extension
        ebounds = empty.check_create_ebounds('test_ebinfile_ebounds.fits')
        self.test_value(ebounds.size(), 1, 'Check number of energy boundaries')
        self.test_value(ebounds.emin().TeV(), 1.0, 'Check minimum energy')
        self.test_value(ebounds.emax().TeV(), 10.0, 'Check maximum energy')

        # Test create_ebounds() method for "ENERGYBINS" extension
        ebounds = empty.check_create_ebounds('test_ebinfile_energybins.fits')
        self.test_value(ebounds.size(), 1, 'Check number of energy boundaries')
        self.test_value(ebounds.emin().TeV(), 1.0, 'Check minimum energy')
        self.test_value(ebounds.emax().TeV(), 10.0, 'Check maximum energy')

        # Test create_ebounds() method for "ENERGIES" extension
        ebounds = empty.check_create_ebounds('test_ebinfile_energies.fits')
        self.test_value(ebounds.size(), 1, 'Check number of energy boundaries')
        self.test_value(ebounds.emin().TeV(), 1.0, 'Check minimum energy')
        self.test_value(ebounds.emax().TeV(), 10.0, 'Check maximum energy')

        # Test create_ebounds() method for invalid extension
        self.test_try('Check create_ebounds() method for invalid extension')
        try:
            empty.check_create_ebounds('test_ebinfile_energies.fits[INVALID]')
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test create_ebounds() method for "BINS" extension
        ebounds = empty.check_create_ebounds('test_ebinfile_bins.fits[BINS]')
        self.test_value(ebounds.size(), 1, 'Check number of energy boundaries')
        self.test_value(ebounds.emin().TeV(), 1.0, 'Check minimum energy')
        self.test_value(ebounds.emax().TeV(), 10.0, 'Check maximum energy')

        # Test create_ebounds() method for "ENGS" extension
        ebounds = empty.check_create_ebounds('test_ebinfile_engs.fits[ENGS]')
        self.test_value(ebounds.size(), 1, 'Check number of energy boundaries')
        self.test_value(ebounds.emin().TeV(), 1.0, 'Check minimum energy')
        self.test_value(ebounds.emax().TeV(), 10.0, 'Check maximum energy')

        # Test create_ebounds() method for invalid extension name
        self.test_try('Check create_ebounds() method for invalid extension name')
        try:
            empty.check_create_ebounds('test_ebinfile_bins.fits')
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test create_ebounds() method for "LIN" and "LOG" methods
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('ebinalg','s','h','LIN','','',''))
        pars.append(gammalib.GApplicationPar('ebinfile','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('enumbins','i','h','1','','',''))
        pars.append(gammalib.GApplicationPar('emin','r','h','1.0','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','10.0','','',''))
        ebounds = cls.check_create_ebounds('')
        self.test_value(ebounds.size(), 1, 'Check number of energy boundaries')
        self.test_value(ebounds.emin().TeV(), 1.0, 'Check minimum energy')
        self.test_value(ebounds.emax().TeV(), 10.0, 'Check maximum energy')
        ebounds = cls.check_create_ebounds('')
        self.test_value(ebounds.size(), 1, 'Check number of energy boundaries')
        self.test_value(ebounds.emin().TeV(), 1.0, 'Check minimum energy')
        self.test_value(ebounds.emax().TeV(), 10.0, 'Check maximum energy')
        cls.pars(cpy_pars)

        # Test create_energies() method for "EBOUNDS" extension
        energies = empty.check_create_energies('test_ebinfile_ebounds.fits')
        self.test_value(energies.size(), 2, 'Check number of energies')
        self.test_value(energies[0].TeV(), 1.0, 'Check first energy')
        self.test_value(energies[1].TeV(), 10.0, 'Check second energy')

        # Test create_energies() method for "ENERGYBINS" extension
        energies = empty.check_create_energies('test_ebinfile_energybins.fits')
        self.test_value(energies.size(), 2, 'Check number of energies')
        self.test_value(energies[0].TeV(), 1.0, 'Check first energy')
        self.test_value(energies[1].TeV(), 10.0, 'Check second energy')

        # Test create_energies() method for "ENERGIES" extension
        energies = empty.check_create_energies('test_ebinfile_energies.fits')
        self.test_value(energies.size(), 2, 'Check number of energies')
        self.test_value(energies[0].TeV(), 1.0, 'Check first energy')
        self.test_value(energies[1].TeV(), 10.0, 'Check second energy')

        # Test create_energies() method for invalid extension
        self.test_try('Check create_energies() method for invalid extension')
        try:
            empty.check_create_energies('test_ebinfile_energies.fits[INVALID]')
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test create_energies() method for "BINS" extension
        energies = empty.check_create_energies('test_ebinfile_bins.fits[BINS]')
        self.test_value(energies.size(), 2, 'Check number of energies')
        self.test_value(energies[0].TeV(), 1.0, 'Check first energy')
        self.test_value(energies[1].TeV(), 10.0, 'Check second energy')

        # Test create_energies() method for "ENGS" extension
        energies = empty.check_create_energies('test_ebinfile_engs.fits[ENGS]')
        self.test_value(energies.size(), 2, 'Check number of energies')
        self.test_value(energies[0].TeV(), 1.0, 'Check first energy')
        self.test_value(energies[1].TeV(), 10.0, 'Check second energy')

        # Test create_energies() method for invalid extension name
        self.test_try('Check create_energies() method for invalid extension name')
        try:
            empty.check_create_energies('test_ebinfile_bins.fits')
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test create_energies() method for "LIN" and "LOG" methods
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('ebinalg','s','h','LIN','','',''))
        pars.append(gammalib.GApplicationPar('ebinfile','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('enumbins','i','h','2','','',''))
        pars.append(gammalib.GApplicationPar('emin','r','h','1.0','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','10.0','','',''))
        energies = cls.check_create_energies('')
        self.test_value(energies.size(), 2, 'Check number of energies')
        self.test_value(energies[0].TeV(), 1.0, 'Check first energy')
        self.test_value(energies[1].TeV(), 10.0, 'Check second energy')
        energies = cls.check_create_energies('')
        self.test_value(energies.size(), 2, 'Check number of energies')
        self.test_value(energies[0].TeV(), 1.0, 'Check first energy')
        self.test_value(energies[1].TeV(), 10.0, 'Check second energy')
        cls.pars(cpy_pars)

        # Test create_map() method
        # TODO: implement

        # Test create_cube() method
        # TODO: implement

        # Test create_cta_obs() method
        # TODO: implement

        # Test require_inobs() method for invalid file name
        self.test_try('Check require_inobs() method for invalid file name')
        try:
            empty.check_require_inobs('NONE')
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test require_inobs_nocube() method for valid file name
        self.test_try('Check require_inobs_nocube() method for event list')
        try:
            empty.check_require_inobs_nocube(self._events)
            self.test_try_success()
        except ValueError:
            self.test_try_failure('Exception thrown')

        # Test require_inobs_nocube() method for invalid file name
        self.test_try('Check require_inobs_nocube() method for counts cube')
        try:
            empty.check_require_inobs_nocube(self._cntcube)
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test get_roi() method for the case that there is no "usepnt"
        # parameter and that there are no valid "ra" and "dec" parameters.
        # In that case an empty ROI should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('ra','r','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('dec','r','h','NONE','','',''))
        roi = cls.check_get_roi()
        self.test_assert(not roi.is_valid(), 'Check that get_roi() for absent '
                        '"usepnt" and invalid "ra" and "dec" parameters returns '
                        'an invalid ROI')
        cls.pars(cpy_pars)

        # Test get_roi() method for the case that there is no "usepnt"
        # parameter and that there is no valid "ra" parameter.
        # In that case an empty ROI should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('ra','r','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('dec','r','h','10.0','','',''))
        roi = cls.check_get_roi()
        self.test_assert(not roi.is_valid(), 'Check that get_roi() for absent '
                        '"usepnt" and an invalid "ra" parameter returns '
                        'an invalid ROI')
        cls.pars(cpy_pars)

        # Test get_roi() method for the case that there is no "usepnt"
        # parameter and that there is no valid "dec" parameter.
        # In that case an empty ROI should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('ra','r','h','10.0','','',''))
        pars.append(gammalib.GApplicationPar('dec','r','h','NONE','','',''))
        roi = cls.check_get_roi()
        self.test_assert(not roi.is_valid(), 'Check that get_roi() for absent '
                        '"usepnt" and an invalid "dec" parameter returns '
                        'an invalid ROI')
        cls.pars(cpy_pars)

        # Test get_roi() method for the case that there is no "usepnt"
        # parameter and that there are valid "ra", "dec" and "rad" parameters.
        # In that case a valid ROI should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('ra','r','h','10.0','','',''))
        pars.append(gammalib.GApplicationPar('dec','r','h','20.0','','',''))
        pars.append(gammalib.GApplicationPar('rad','r','h','1.0','','',''))
        roi = cls.check_get_roi()
        self.test_assert(roi.is_valid(), 'Check that get_roi() for absent '
                        '"usepnt" and valid "ra", "dec" and "rad" parameters '
                        'returns a valid ROI')
        self.test_value(roi.centre().dir().ra_deg(), 10.0, 'Check Right Ascension')
        self.test_value(roi.centre().dir().dec_deg(), 20.0, 'Check Declination')
        self.test_value(roi.radius(), 1.0, 'Check radius')
        cls.pars(cpy_pars)

        # Test get_roi() method for the case that there is a "usepnt"
        # parameter and that there is a valid "rad" parameter.
        # In that case a valid ROI should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('usepnt','b','h','yes','','',''))
        pars.append(gammalib.GApplicationPar('rad','r','h','1.0','','',''))
        roi = cls.check_get_roi()
        self.test_assert(roi.is_valid(), 'Check that get_roi() for present '
                        '"usepnt" and valid "rad" parameter returns a valid ROI')
        self.test_value(roi.centre().dir().ra_deg(), 0.0, 'Check Right Ascension')
        self.test_value(roi.centre().dir().dec_deg(), 0.0, 'Check Declination')
        self.test_value(roi.radius(), 1.0, 'Check radius')
        cls.pars(cpy_pars)

        # Test get_ebounds() method for the case that both "emin" and "emax"
        # parameters are invalid. In that case an empty energy boundary should
        # be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('emin','r','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','NONE','','',''))
        ebounds = cls.check_get_ebounds()
        self.test_value(ebounds.size(), 0, 'Check that get_ebounds() for '
                        'invalid "emin" and "emax" parameters returns empty '
                        'energy boundaries')
        cls.pars(cpy_pars)

        # Test get_ebounds() method for the case that the "emin" parameter is
        # invalid. In that case an empty energy boundary should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('emin','r','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','100.0','','',''))
        ebounds = cls.check_get_ebounds()
        self.test_value(ebounds.size(), 0, 'Check that get_ebounds() for '
                        'invalid "emin" parameter returns empty energy '
                        'boundaries')
        cls.pars(cpy_pars)

        # Test get_ebounds() method for the case that the "emax" parameter is
        # invalid. In that case an empty energy boundary should be returned.
        test, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('emin','r','h','1.0','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','NONE','','',''))
        ebounds = test.check_get_ebounds()
        self.test_value(ebounds.size(), 0, 'Check that get_ebounds() for '
                        'invalid "emax" parameter returns empty energy '
                        'boundaries')
        test.pars(cpy_pars)

        # Test get_ebounds() method for the case of valid "emin" and "emax"
        # parameters. In that case a single energy boundary should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('emin','r','h','1.0','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','100.0','','',''))
        ebounds = cls.check_get_ebounds()
        self.test_value(ebounds.size(), 1, 'Check that get_ebounds() for '
                        'valid "emin" and "emax" parameters returns empty one '
                        'energy boundary')
        self.test_value(ebounds.emin().TeV(), 1.0, 'Check minimum energy')
        self.test_value(ebounds.emax().TeV(), 100.0, 'Check maximum energy')
        cls.pars(cpy_pars)

        # Test get_gti() method for the case that both "tmin" and "tmax"
        # parameters are invalid. In that case an empty Good Time Interval
        # should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('tmin','t','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('tmax','t','h','NONE','','',''))
        gti = cls.check_get_gti()
        self.test_value(gti.size(), 0, 'Check that get_gti() for invalid "tmin" '
                        'and "tmax" parameters returns empty GTI')
        cls.pars(cpy_pars)

        # Test get_gti() method for the case that the "tmin" parameter is
        # invalid. In that case an empty Good Time Interval should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('tmin','t','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('tmax','t','h','2005-10-08T14:58:26','','',''))
        gti = cls.check_get_gti()
        self.test_value(gti.size(), 0, 'Check that get_gti() for an invalid '
                        '"tmin" parameter returns empty GTI')
        cls.pars(cpy_pars)

        # Test get_gti() method for the case that the "tmax" parameter is
        # invalid. In that case an empty Good Time Interval should be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('tmin','t','h','2005-10-08T14:30:25','','',''))
        pars.append(gammalib.GApplicationPar('tmax','t','h','NONE','','',''))
        gti = cls.check_get_gti()
        self.test_value(gti.size(), 0, 'Check that get_gti() for an invalid '
                        '"tmax" parameter returns empty GTI')
        cls.pars(cpy_pars)

        # Test get_gti() method for the case that both "tmin" and "tmax"
        # parameters are valid. In that case a valid Good Time Interval should
        # be returned.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('tmin','t','h','2005-10-08T14:30:25','','',''))
        pars.append(gammalib.GApplicationPar('tmax','t','h','2005-10-08T14:58:26','','',''))
        gti = cls.check_get_gti()
        self.test_value(gti.size(), 1, 'Check that get_gti() for valid "tmin" '
                        'and "tmax" parameters returns one GTI')
        self.test_value(gti.tstart().utc(), '2005-10-08T14:30:25', 'Check "tmin"')
        self.test_value(gti.tstop().utc(), '2005-10-08T14:58:26', 'Check "tmax"')
        cls.pars(cpy_pars)

        # Test get_pointing() method
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('ra','r','h','10.0','','',''))
        pars.append(gammalib.GApplicationPar('dec','r','h','20.0','','',''))
        pnt = cls.check_get_pointing()
        self.test_value(pnt.dir().ra_deg(), 10.0, 'Check Right Ascension')
        self.test_value(pnt.dir().dec_deg(), 20.0, 'Check Declination')
        cls.pars(cpy_pars)

        # Test get_skydir() method
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('ra','r','h','10.0','','',''))
        pars.append(gammalib.GApplicationPar('dec','r','h','20.0','','',''))
        dir = cls.check_get_skydir()
        self.test_value(dir.ra_deg(), 10.0, 'Check Right Ascension')
        self.test_value(dir.dec_deg(), 20.0, 'Check Declination')
        cls.pars(cpy_pars)

        # Test set_outfile_name() method for "fix" filename, which should
        # return the prefixed basename.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('prefix','s','h','pre_','','',''))
        filename = cls.check_set_outfile_name('fix')
        self.test_value(filename, 'pre_fix', 'Check "pre_fix"')
        cls.pars(cpy_pars)

        # Test set_outfile_name() method for "/path/to/fix" filename, which
        # should return the prefixed basename.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('prefix','s','h','pre_','','',''))
        filename = cls.check_set_outfile_name('/path/to/fix')
        self.test_value(filename, 'pre_fix', 'Check "pre_fix"')
        cls.pars(cpy_pars)

        # Test set_outfile_name() method for "/path/to/fix.gz" filename, which
        # should return the prefixed basename without the ".gz"
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('prefix','s','h','pre_','','',''))
        filename = cls.check_set_outfile_name('/path/to/fix.gz')
        self.test_value(filename, 'pre_fix', 'Check "pre_fix"')
        cls.pars(cpy_pars)

        # Test set_outfile_name() method for empty filename, which should
        # return an empty filename.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('prefix','s','h','pre_','','',''))
        filename = cls.check_set_outfile_name('')
        self.test_value(filename, '', 'Check on empty string')
        cls.pars(cpy_pars)

        # Test is_stacked() method for the case that "enumbins" is zero. In
        # that case the is_stacked() method should return False.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('emin','r','h','1.0','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','100.0','','',''))
        pars.append(gammalib.GApplicationPar('enumbins','i','h','0','','',''))
        self.test_assert(not cls.check_is_stacked(), 'Check if is_stacked() '
                         'returns False if there are no energy bins')
        cls.pars(cpy_pars)

        # Test is_stacked() method for the case that "enumbins" is larger
        # than zero. In that case the is_stacked() method should return True
        # and query a bunch of parameters.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('emin','r','h','1.0','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','100.0','','',''))
        pars.append(gammalib.GApplicationPar('enumbins','i','h','10','','',''))
        pars.append(gammalib.GApplicationPar('coordsys','s','h','CEL','','',''))
        pars.append(gammalib.GApplicationPar('proj','s','h','TAN','','',''))
        pars.append(gammalib.GApplicationPar('xref','r','h','11.1','','',''))
        pars.append(gammalib.GApplicationPar('yref','r','h','12.1','','',''))
        pars.append(gammalib.GApplicationPar('nxpix','i','h','100','','',''))
        pars.append(gammalib.GApplicationPar('nypix','i','h','100','','',''))
        pars.append(gammalib.GApplicationPar('binsz','r','h','0.02','','',''))
        self.test_assert(cls.check_is_stacked(), 'Check if is_stacked() '
                         'returns True if there are energy bins')
        cls.pars(cpy_pars)

        # Test is_onoff() method for the case that method='NONE', i.e. the
        # method is not 'ONOFF'. In that case the is_onoff() method should
        # return False.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('method','s','h','NONE','','',''))
        self.test_assert(not cls.check_is_onoff(),
             'Check if is_onoff() returns False if method is not "ONOFF"')
        cls.pars(cpy_pars)

        # Test is_onoff() method for the case that method='ONOFF',
        # srcshape='CIRCLE' and bkgmethod='REFLECTED'. In that case the
        # is_onoff() method should return True and a bunch of application
        # parameters are queried.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('method','s','h','ONOFF','','',''))
        pars.append(gammalib.GApplicationPar('inexclusion','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('emin','r','h','0.1','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','100.0','','',''))
        pars.append(gammalib.GApplicationPar('enumbins','i','h','10','','',''))
        pars.append(gammalib.GApplicationPar('coordsys','s','h','CEL','','',''))
        pars.append(gammalib.GApplicationPar('xref','r','h','11.1','','',''))
        pars.append(gammalib.GApplicationPar('yref','r','h','12.1','','',''))
        pars.append(gammalib.GApplicationPar('srcshape','s','h','CIRCLE','','',''))
        pars.append(gammalib.GApplicationPar('rad','r','h','1.1','','',''))
        pars.append(gammalib.GApplicationPar('bkgmethod','s','h','REFLECTED','','',''))
        pars.append(gammalib.GApplicationPar('bkgregmin','i','h','3','','',''))
        pars.append(gammalib.GApplicationPar('maxoffset','r','h','0.1','','',''))
        pars.append(gammalib.GApplicationPar('etruemin','r','h','0.1','','',''))
        pars.append(gammalib.GApplicationPar('etruemax','r','h','100.0','','',''))
        pars.append(gammalib.GApplicationPar('etruebins','i','h','10','','',''))
        self.test_assert(cls.check_is_onoff(),
             'Check if is_onoff() returns True if method is "ONOFF"')
        cls.pars(cpy_pars)

        # Test is_onoff() method for the case that method='ONOFF',
        # srcshape='CIRCLE' and bkgmethod='REFLECTED'. In that case the
        # is_onoff() method should return True and a bunch of application
        # parameters are queried, except of rad and bkgregmin.
        cls, pars, cpy_pars = self._setup_cscript()
        pars.append(gammalib.GApplicationPar('method','s','h','ONOFF','','',''))
        pars.append(gammalib.GApplicationPar('inexclusion','f','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('emin','r','h','0.1','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','100.0','','',''))
        pars.append(gammalib.GApplicationPar('enumbins','i','h','10','','',''))
        pars.append(gammalib.GApplicationPar('coordsys','s','h','CEL','','',''))
        pars.append(gammalib.GApplicationPar('xref','r','h','11.1','','',''))
        pars.append(gammalib.GApplicationPar('yref','r','h','12.1','','',''))
        pars.append(gammalib.GApplicationPar('srcshape','s','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('bkgmethod','s','h','NONE','','',''))
        pars.append(gammalib.GApplicationPar('maxoffset','r','h','0.1','','',''))
        pars.append(gammalib.GApplicationPar('etruemin','r','h','0.1','','',''))
        pars.append(gammalib.GApplicationPar('etruemax','r','h','100.0','','',''))
        pars.append(gammalib.GApplicationPar('etruebins','i','h','10','','',''))
        self.test_assert(cls.check_is_onoff(),
             'Check if is_onoff() returns True if method is "ONOFF"')
        cls.pars(cpy_pars)

        # Test set_response() method
        # TODO: implement

        # Test set_edisp() method
        # TODO: implement

        # Test restore_edisp() method for invalid vector size
        self.test_try('Check restore_edisp() method for invalid vector size')
        try:
            empty.check_restore_edisp()
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test set_obs_response() method with 'NONE' files. If energy dispersion
        # is requested then an exception will be thrown.
        empty.check_set_obs_response('NONE', 'NONE', 'NONE', 'NONE', 'no')
        empty.check_set_obs_response('NONE', 'NONE', 'NONE', 'NONE', 'yes')
        self.test_try('Check set_obs_response() method for file names')
        try:
            empty.check_set_obs_response(self._datadir+'/crab_expcube.fits',
                                         self._datadir+'/crab_psfcube.fits',
                                         self._datadir+'/crab_bkgcube.fits',
                                         'NONE', 'yes')
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()
        empty.check_set_obs_response(self._datadir+'/crab_expcube.fits',
                                     self._datadir+'/crab_psfcube.fits',
                                     self._datadir+'/crab_bkgcube.fits',
                                     self._datadir+'/crab_edispcube.fits',
                                     'yes')

        # Test get_mean_pointing() method for empty observation container
        self.test_try('Check get_mean_pointing() method for empty observation '
                      'container')
        try:
            obs = gammalib.GObservations()
            empty.check_get_mean_pointing(obs)
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test get_mean_pointing() method for observation container with LAT obs
        self.test_try('Check get_mean_pointing() method for empty observation '
                      'container')
        try:
            lat = gammalib.GLATObservation()
            obs = gammalib.GObservations()
            obs.append(lat)
            empty.check_get_mean_pointing(obs)
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test get_current_rss() method
        empty.check_get_current_rss()

        # Test get_obs_header() method
        # TODO: implement

        # Test insert_energy_boundaries() method
        # TODO: implement

        # Test cube_layer_usage() method
        # TODO: implement

        # Check save_event_list() method
        empty.check_save_event_list('test_event_list_input1.fits', 'EVENTS',
                                    'GTI', 'test_event_list_output1.fits')
        self._test_cscript_save_event_list('test_event_list_output1.fits',
                                           ['Primary','EVENTS','GTI'])
        empty.check_save_event_list('test_event_list_input2.fits', 'EVENTS',
                                    'GTI', 'test_event_list_output2.fits[EVT]')
        self._test_cscript_save_event_list('test_event_list_output2.fits',
                                           ['Primary','EVT','GTI'])
        empty.check_save_event_list('test_event_list_input3.fits', 'EVENTS',
                                    'GTI', 'test_event_list_output3.fits[EVT;TIME]')
        self._test_cscript_save_event_list('test_event_list_output3.fits',
                                           ['Primary','EVT','TIME'])
        empty.check_save_event_list('test_event_list_input4.fits', 'EVENTS',
                                    'GTI', 'test_event_list_output4.fits[EVT;]')
        self._test_cscript_save_event_list('test_event_list_output4.fits',
                                           ['Primary','EVT','GTI'])
        empty.check_save_event_list('test_event_list_input5.fits', 'EVENTS',
                                    'GTI', 'test_event_list_output5.fits[;TIME]')
        self._test_cscript_save_event_list('test_event_list_output5.fits',
                                           ['Primary','EVENTS','TIME'])
        empty.check_save_event_list('test_event_list_input6.fits', 'XEVENTS',
                                    'XGTI', 'test_event_list_output6.fits[EVT;TIME]')
        self._test_cscript_save_event_list('test_event_list_output6.fits',
                                           ['Primary','EVT','TIME','EVENTS','GTI'])

        # Test get_gtiname() method
        # TODO: implement

        # Test warn_too_few_energies() method
        energies_empty = gammalib.GEnergies()
        energies_2     = gammalib.GEnergies(2, gammalib.GEnergy(1.0,'TeV'),
                                               gammalib.GEnergy(10.0,'TeV'))
        energies_100   = gammalib.GEnergies(100, gammalib.GEnergy(1.0,'TeV'),
                                                 gammalib.GEnergy(10.0,'TeV'))
        self.test_value(len(empty.check_warn_too_few_energies(energies_empty)), 0,
                        'Check length of warning string for too few energies')
        self.test_value(len(empty.check_warn_too_few_energies(energies_2)), 253,
                        'Check length of warning string for too few energies')
        self.test_value(len(empty.check_warn_too_few_energies(energies_100)), 0,
                        'Check length of warning string for too few energies')

        # Test warn_xml_suffix() method
        self.test_value(len(empty.check_warn_xml_suffix('test.xml')), 0,
                        'Check xml suffix warning for "test.xml" file')
        self.test_value(len(empty.check_warn_xml_suffix('test.fits')), 235,
                        'Check xml suffix warning for "test.fits" file')

        # Return
        return

    # Test check_save_event_list() method result
    def _test_cscript_save_event_list(self, filename, hdunames):
        """
        Test check_save_event_list() method result
        """
        # Open FITS file
        fits = gammalib.GFits(filename)

        # Check number of HDUs
        self.test_value(fits.size(), len(hdunames), 'save_event_list(): '
                        'Check number of FITS HDUs in file %s' % filename)

        # Check HDU extension names
        for i, hdu in enumerate(fits):
            self.test_value(hdu.extname(), hdunames[i], 'save_event_list(): '
                            'Check FITS HDU extension name in file %s' % filename)

        # Return
        return

    # Test ctobservation base class
    def _test_ctobservation(self):
        """
        Test ctobservation base class
        """
        # Allocate empty
        empty = ctobservation_test()

        # Test first_unbinned_observation() method
        self.test_assert(empty.check_first_unbinned_observation() is None,
                         'Check first_unbinned_observation() method')

        # Test next_unbinned_observation() method
        self.test_assert(empty.check_next_unbinned_observation() is None,
                         'Check next_unbinned_observation() method')

        # Test set_obs_statistic() method for unbinned observation
        self.test_value(empty.check_set_obs_statistic_unbinned('CSTAT'),
                        'CSTAT',
                        'Check set_obs_statistic("CSTAT") for unbinned')
        self.test_value(empty.check_set_obs_statistic_unbinned('WSTAT'),
                        'CSTAT',
                        'Check set_obs_statistic("WSTAT") for unbinned')
        self.test_value(empty.check_set_obs_statistic_unbinned('CHI2'),
                        'CSTAT',
                        'Check set_obs_statistic("CHI2") for unbinned')
        self.test_value(empty.check_set_obs_statistic_unbinned('DEFAULT'),
                        'CSTAT',
                        'Check set_obs_statistic("DEFAULT") for unbinned')

        # Test set_obs_statistic() method for binned observation
        self.test_value(empty.check_set_obs_statistic_binned('CSTAT'),
                        'CSTAT',
                        'Check set_obs_statistic("CSTAT") for binned')
        self.test_value(empty.check_set_obs_statistic_binned('WSTAT'),
                        'CSTAT',
                        'Check set_obs_statistic("WSTAT") for binned')
        self.test_value(empty.check_set_obs_statistic_binned('CHI2'),
                        'CHI2',
                        'Check set_obs_statistic("CHI2") for binned')
        self.test_value(empty.check_set_obs_statistic_binned('DEFAULT'),
                        'CSTAT',
                        'Check set_obs_statistic("DEFAULT") for binned')

        # Test set_obs_statistic() method for On/Off observation
        self.test_value(empty.check_set_obs_statistic_onoff('CSTAT'),
                        'CSTAT',
                        'Check set_obs_statistic("CSTAT") for On/Off')
        self.test_value(empty.check_set_obs_statistic_onoff('WSTAT'),
                        'WSTAT',
                        'Check set_obs_statistic("WSTAT") for On/Off')
        self.test_value(empty.check_set_obs_statistic_onoff('CHI2'),
                        'CSTAT',
                        'Check set_obs_statistic("CHI2") for On/Off')
        self.test_value(empty.check_set_obs_statistic_onoff('DEFAULT'),
                        'CSTAT',
                        'Check set_obs_statistic("DEFAULT") for On/Off')

        # Test set_obs_bounds() method
        cls, pars, cpy_pars = self._setup_ctobservation_test()
        pars.append(gammalib.GApplicationPar('ra','r','h','10.0','','',''))
        pars.append(gammalib.GApplicationPar('dec','r','h','20.0','','',''))
        pars.append(gammalib.GApplicationPar('rad','r','h','3.0','','',''))
        pars.append(gammalib.GApplicationPar('emin','r','h','1.0','','',''))
        pars.append(gammalib.GApplicationPar('emax','r','h','100.0','','',''))
        pars.append(gammalib.GApplicationPar('tmin','t','h','2005-10-08T14:30:25','','',''))
        pars.append(gammalib.GApplicationPar('tmax','t','h','2005-10-08T14:58:26','','',''))
        obs = cls.check_set_obs_bounds()
        self.test_value(obs.size(), 1, 'Check number of observations')
        self.test_value(obs[0].events().gti().size(), 1,
                        'Check number of GTIs')
        self.test_value(obs[0].events().gti().tstart().utc(), '2005-10-08T14:30:25',
                        'Check start time')
        self.test_value(obs[0].events().gti().tstop().utc(), '2005-10-08T14:58:26',
                        'Check stop time')
        self.test_value(obs[0].events().ebounds().size(), 1,
                        'Check number of energy boundaries')
        self.test_value(obs[0].events().ebounds().emin().TeV(), 1.0,
                        'Check minimum energy')
        self.test_value(obs[0].events().ebounds().emax().TeV(), 100.0,
                        'Check maximum energy')
        self.test_value(obs[0].events().roi().centre().dir().ra_deg(), 10.0,
                        'Check RoI centre Right Ascension')
        self.test_value(obs[0].events().roi().centre().dir().dec_deg(), 20.0,
                        'Check RoI centre Declination')
        self.test_value(obs[0].events().roi().radius(), 3.0,
                        'Check RoI radius')
        cls.pars(cpy_pars)

        # Test read_ogip_keywords() method for empty header
        hdu = empty.check_ogip_keywords(gammalib.GFitsBinTable())
        self.test_value(hdu.string('TELESCOP'), '', 'Check "TELESCOP" keyword')
        self.test_value(hdu.string('DATE-OBS'), '2010-01-01', 'Check "DATE-OBS" keyword')
        self.test_value(hdu.string('TIME-OBS'), '00:00:00', 'Check "TIME-OBS" keyword')
        self.test_value(hdu.string('DATE-END'), '2010-01-01', 'Check "DATE-END" keyword')
        self.test_value(hdu.string('TIME-END'), '00:00:00', 'Check "TIME-END" keyword')
        self.test_value(hdu.real('TELAPSE'), 0.0, 'Check "TELAPSE" keyword')
        self.test_value(hdu.real('ONTIME'), 0.0, 'Check "ONTIME" keyword')
        self.test_value(hdu.real('LIVETIME'), 0.0, 'Check "LIVETIME" keyword')
        self.test_value(hdu.real('EXPOSURE'), 0.0, 'Check "EXPOSURE" keyword')
        self.test_value(hdu.real('DEADC'), 1.0, 'Check "DEADC" keyword')

        # Test read_ogip_keywords() method for filled header
        table = gammalib.GFitsBinTable()
        table.card('TELESCOP', 'CTA', 'Comment')
        table.card('DATE-OBS', '2018-01-01', 'Comment')
        table.card('TIME-OBS', '01:02:03', 'Comment')
        table.card('DATE-END', '2018-02-05', 'Comment')
        table.card('TIME-END', '21:12:05', 'Comment')
        table.card('TELAPSE', 1234.0, 'Comment')
        table.card('ONTIME', 2345.0, 'Comment')
        table.card('LIVETIME', 2210.0, 'Comment')
        table.card('EXPOSURE', 3001.0, 'Comment')
        hdu = empty.check_ogip_keywords(table)
        self.test_value(hdu.string('TELESCOP'), 'CTA', 'Check "TELESCOP" keyword')
        self.test_value(hdu.string('DATE-OBS'), '2018-01-01', 'Check "DATE-OBS" keyword')
        self.test_value(hdu.string('TIME-OBS'), '01:02:03', 'Check "TIME-OBS" keyword')
        self.test_value(hdu.string('DATE-END'), '2018-02-05', 'Check "DATE-END" keyword')
        self.test_value(hdu.string('TIME-END'), '21:12:05', 'Check "TIME-END" keyword')
        self.test_value(hdu.real('TELAPSE'), 1234.0, 'Check "TELAPSE" keyword')
        self.test_value(hdu.real('ONTIME'), 2345.0, 'Check "ONTIME" keyword')
        self.test_value(hdu.real('LIVETIME'), 2210.0, 'Check "LIVETIME" keyword')
        self.test_value(hdu.real('EXPOSURE'), 3001.0, 'Check "EXPOSURE" keyword')
        self.test_value(hdu.real('DEADC'), 2210.0/2345.0, 'Check "DEADC" keyword')

        # Return
        return

    # Test ctlikelihood base class
    def _test_ctlikelihood(self):
        """
        Test ctlikelihood base class
        """
        # Allocate empty
        empty = ctlikelihood_test()

        # Test evaluate() method for minimum value violation
        self.test_try('Check evaluate() for minimum value violation')
        try:
            empty.check_evaluate(0.0)
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Test evaluate() method for maximum value violation
        self.test_try('Check evaluate() for maximum value violation')
        try:
            empty.check_evaluate(4.0)
            self.test_try_failure('Exception not thrown')
        except ValueError:
            self.test_try_success()

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csbkgmodel script.
#
# Copyright (C) 2019 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ================================ #
# Test class for csbkgmodel script #
# ================================ #
class Test(test):
    """
    Test class for csbkgmodel script

    This test class makes unit tests for the csbkgmodel script by using it
    from the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set test datasets and parameters
        self._myevents1 = self._datadir + '/crab_offaxis1.fits'
        self._myevents2 = self._datadir + '/crab_offaxis2.fits'
        self._lookup    = self._datadir + '/hess_bkg_lookup.fits'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('csbkgmodel')

        # Append tests
        self.append(self._test_cmd, 'Test csbkgmodel on command line')
        self.append(self._test_python, 'Test csbkgmodel from Python')
        self.append(self._test_pickeling, 'Test csbkgmodel pickeling')

        # Return
        return

    # Test csbkgmodel on command line
    def _test_cmd(self):
        """
        Test csbkgmodel on the command line.
        """
        # Set script name
        csbkgmodel = self._script('csbkgmodel')

        # Setup csbkgmodel command
        cmd = csbkgmodel+' inobs="'+self._events+'"'+ \
                         ' caldb="'+self._caldb+'" irf="'+self._irf+'"' \
                         ' instrument=CTA spatial=GAUSS gradient=yes'+ \
                         ' spectral=NODES ebinalg=LOG emin=1.0 emax=100.0'+ \
                         ' enumbins=8 runwise=yes rad=2.0'+ \
                         ' outmodel="csbkgmodel_cmd1.xml"'+ \
                         ' logfile="csbkgmodel_cmd1.log" chatter=4'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check background model
        self._check_bkg_model('csbkgmodel_cmd1.xml')

        # Setup csbkgmodel command
        cmd = csbkgmodel+' inobs="events_that_do_not_exist.fits"'+ \
                         ' caldb="'+self._caldb+'" irf="'+self._irf+'"' \
                         ' instrument=CTA spatial=GAUSS gradient=yes'+ \
                         ' spectral=NODES ebinalg=LOG emin=1.0 emax=100.0'+ \
                         ' enumbins=8 runwise=yes rad=2.0'+ \
                         ' outmodel="csbkgmodel_cmd2.xml"' + \
                         ' logfile="csbkgmodel_cmd2.log" chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csbkgmodel --help
        self._check_help(csbkgmodel)

        # Return
        return

    # Test csbkgmodel from Python
    def _test_python(self):
        """
        Test csbkgmodel from Python
        """
        # Set-up csbkgmodel
        bkgmodel = cscripts.csbkgmodel()
        bkgmodel['inobs']      = self._events
        bkgmodel['caldb']      = self._caldb
        bkgmodel['irf']        = self._irf
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'GAUSS'
        bkgmodel['gradient']   = True
        bkgmodel['spectral']   = 'NODES'
        bkgmodel['ebinalg']    = 'LOG'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['runwise']    = True
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 2
        bkgmodel['outmodel']   = 'csbkgmodel_py1.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py1.log'

        # Run csbkgmodel script and save background model
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.run()
        bkgmodel.save()

        # Check background model
        self._check_bkg_model('csbkgmodel_py1.xml')

        # Now test without gradient, power law and not runwise
        bkgmodel = cscripts.csbkgmodel()
        bkgmodel['inobs']      = self._events
        bkgmodel['caldb']      = self._caldb
        bkgmodel['irf']        = self._irf
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'GAUSS'
        bkgmodel['gradient']   = False
        bkgmodel['spectral']   = 'PLAW'
        bkgmodel['ebinalg']    = 'LOG'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['runwise']    = False
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 3
        bkgmodel['outmodel']   = 'csbkgmodel_py2.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py2.log'

        # Execute csbkgmodel script
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.execute()

        # Check background model
        self._check_bkg_model('csbkgmodel_py2.xml')

        # Now test AEFF model
        bkgmodel = cscripts.csbkgmodel()
        bkgmodel['inobs']      = self._events
        bkgmodel['caldb']      = self._caldb
        bkgmodel['irf']        = self._irf
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'AEFF'
        bkgmodel['gradient']   = False
        bkgmodel['spectral']   = 'PLAW'
        bkgmodel['ebinalg']    = 'LOG'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['runwise']    = False
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 4
        bkgmodel['outmodel']   = 'csbkgmodel_py3.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py3.log'

        # Execute csbkgmodel script
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.execute()

        # Check background model
        self._check_bkg_model('csbkgmodel_py3.xml')

        # Now test IRF model
        bkgmodel = cscripts.csbkgmodel()
        bkgmodel['inobs']      = self._events
        bkgmodel['caldb']      = self._caldb
        bkgmodel['irf']        = self._irf
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'IRF'
        bkgmodel['gradient']   = False
        bkgmodel['spectral']   = 'PLAW'
        bkgmodel['ebinalg']    = 'LOG'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['runwise']    = False
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 4
        bkgmodel['outmodel']   = 'csbkgmodel_py4.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py4.log'

        # Execute csbkgmodel script
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.execute()

        # Check background model
        self._check_bkg_model('csbkgmodel_py4.xml')

        # Test with multiple input observations
        obs = gammalib.GObservations()
        for s, events in enumerate([self._myevents1, self._myevents2]):
            run = gammalib.GCTAObservation(events)
            run.id(str(s + 1))
            run.response(self._irf, gammalib.GCaldb('cta', self._caldb))
            obs.append(run)

        # Set-up csbkgmodel
        bkgmodel = cscripts.csbkgmodel(obs)
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'GAUSS'
        bkgmodel['gradient']   = True
        bkgmodel['spectral']   = 'NODES'
        bkgmodel['ebinalg']    = 'POW'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['ebingamma']  = 1.1
        bkgmodel['runwise']    = True
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 2
        bkgmodel['outmodel']   = 'csbkgmodel_py5.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py5.log'

        # Execute csbkgmodel script
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.execute()

        # Check background model
        self._check_bkg_model('csbkgmodel_py5.xml', nmodels=2)

        # Test GAUSS(E) spatial model
        bkgmodel = cscripts.csbkgmodel()
        bkgmodel['inobs']      = self._events
        bkgmodel['caldb']      = self._caldb
        bkgmodel['irf']        = self._irf
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'GAUSS(E)'
        bkgmodel['snumbins']   = 2
        bkgmodel['smin']       = 1.0
        bkgmodel['smax']       = 10.0
        bkgmodel['gradient']   = True
        bkgmodel['spectral']   = 'NODES'
        bkgmodel['ebinalg']    = 'LOG'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['runwise']    = True
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 2
        bkgmodel['outmodel']   = 'csbkgmodel_py6.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py6.log'

        # Run csbkgmodel script and save background model
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.execute()

        # Check background model
        self._check_bkg_model('csbkgmodel_py6.xml')

        # Test LOOKUP spatial model
        bkgmodel = cscripts.csbkgmodel()
        bkgmodel['inobs']      = self._events
        bkgmodel['caldb']      = self._caldb
        bkgmodel['irf']        = self._irf
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'LOOKUP'
        bkgmodel['slufile']    = self._lookup
        bkgmodel['gradient']   = True
        bkgmodel['spectral']   = 'NODES'
        bkgmodel['ebinalg']    = 'LOG'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['runwise']    = True
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 2
        bkgmodel['outmodel']   = 'csbkgmodel_py7.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py7.log'

        # Run csbkgmodel script and save background model
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.execute()

        # Check background model
        self._check_bkg_model('csbkgmodel_py7.xml')

        # Test PROFILE spatial model
        bkgmodel = cscripts.csbkgmodel()
        bkgmodel['inobs']      = self._events
        bkgmodel['caldb']      = self._caldb
        bkgmodel['irf']        = self._irf
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'PROFILE'
        bkgmodel['gradient']   = True
        bkgmodel['spectral']   = 'NODES'
        bkgmodel['ebinalg']    = 'LOG'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['runwise']    = True
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 2
        bkgmodel['outmodel']   = 'csbkgmodel_py8.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py8.log'

        # Run csbkgmodel script and save background model
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.execute()

        # Check background model
        self._check_bkg_model('csbkgmodel_py8.xml')

        # Test POLYNOM spatial model
        bkgmodel = cscripts.csbkgmodel()
        bkgmodel['inobs']      = self._events
        bkgmodel['caldb']      = self._caldb
        bkgmodel['irf']        = self._irf
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'POLYNOM'
        bkgmodel['gradient']   = True
        bkgmodel['spectral']   = 'NODES'
        bkgmodel['ebinalg']    = 'LOG'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['runwise']    = True
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 2
        bkgmodel['outmodel']   = 'csbkgmodel_py9.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py9.log'

        # Run csbkgmodel script and save background model
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.execute()

        # Check background model
        self._check_bkg_model('csbkgmodel_py9.xml')

        # Check with data from multiple instruments
        new_inst = 'INST2'
        obs_multi_inst = gammalib.GObservations(obs)
        for run in obs_multi_inst:
            run.instrument(new_inst)
        obs_multi_inst.extend(obs)

        # Set-up csbkgmodel
        bkgmodel = cscripts.csbkgmodel(obs_multi_inst)
        bkgmodel['instrument'] = new_inst
        bkgmodel['spatial']    = 'GAUSS'
        bkgmodel['gradient']   = True
        bkgmodel['spectral']   = 'NODES'
        bkgmodel['ebinalg']    = 'POW'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['ebingamma']  = 1.1
        bkgmodel['runwise']    = True
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 2
        bkgmodel['outmodel']   = 'csbkgmodel_py10.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py10.log'

        # Execute csbkgmodel script
        bkgmodel.logFileOpen()   # Make sure we get a log file
        bkgmodel.execute()

        # Check background model
        self._check_bkg_model('csbkgmodel_py10.xml', nmodels=obs.size())

        # Return
        return

    # Test csbkgmodel pickeling
    def _test_pickeling(self):
        """
        Test csbkgmodel pickeling
        """
        # Perform pickeling tests of empty class
        self._pickeling(cscripts.csbkgmodel())

        # Set-up csbkgmodel
        bkgmodel = cscripts.csbkgmodel()
        bkgmodel['inobs']      = self._events
        bkgmodel['caldb']      = self._caldb
        bkgmodel['irf']        = self._irf
        bkgmodel['instrument'] = 'CTA'
        bkgmodel['spatial']    = 'GAUSS'
        bkgmodel['gradient']   = True
        bkgmodel['spectral']   = 'NODES'
        bkgmodel['ebinalg']    = 'LOG'
        bkgmodel['emin']       = 1.0
        bkgmodel['emax']       = 100.0
        bkgmodel['enumbins']   = 8
        bkgmodel['runwise']    = True
        bkgmodel['rad']        = 2.0
        bkgmodel['chatter']    = 2
        bkgmodel['outmodel']   = 'csbkgmodel_py1_pickle.xml'
        bkgmodel['logfile']    = 'csbkgmodel_py1_pickle.log'

        # Perform pickeling tests of filled class
        obj = self._pickeling(bkgmodel)

        # Run csbkgmodel script and save background model
        obj.logFileOpen()   # Make sure we get a log file
        obj.run()
        obj.save()

        # Check background model
        self._check_bkg_model('csbkgmodel_py1_pickle.xml')

        # Return
        return

    # Check background model
    def _check_bkg_model(self, filename, nmodels=1):
        """
        Check background model
        """
        # Load model
        models = gammalib.GModels(filename)

        # Check model size
        self.test_value(models.size(), nmodels,
                        'Check for %d model(s) in XML file' % nmodels)

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctedispcube tool.
#
# Copyright (C) 2016-2020 Maria Haupt
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from testing import test


# =============================== #
# Test class for ctedispcube tool #
# =============================== #
class Test(test):
    """
    Test class for ctedispcube tool
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctedispcube')

        # Append tests
        self.append(self._test_cmd, 'Test ctedispcube on command line')
        self.append(self._test_python, 'Test ctedispcube from Python')

        # Return
        return

    # Test ctedispcube on command line
    def _test_cmd(self):
        """
        Test ctedispcube on the command line
        """
        # Set tool name
        ctedispcube = self._tool('ctedispcube')

        # Setup ctedispcube command
        cmd = ctedispcube+' inobs="'+self._events+'"'+ \
                          ' incube="NONE"'+ \
                          ' outcube="ctedispcube_cmd1.fits"'+ \
                          ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                          ' ebinalg="LOG" emin=1.0 emax=100.0 enumbins=5'+ \
                          ' nxpix=10 nypix=10 binsz=0.4'+ \
                          ' coordsys="CEL" proj="CAR" xref=83.63 yref=22.01'+ \
                          ' migramax=2.0 migrabins=10'+ \
                          ' logfile="ctedispcube_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('ctedispcube_cmd1.fits')

        # Setup ctedispcube command
        cmd = ctedispcube+' inobs="events_that_do_not_exist.fits"'+ \
                          ' incube="NONE"'+ \
                          ' outcube="ctedispcube_cmd2.fits"'+ \
                          ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                          ' ebinalg="LOG" emin=1.0 emax=100.0 enumbins=5'+ \
                          ' nxpix=10 nypix=10 binsz=0.4'+ \
                          ' coordsys="CEL" proj="CAR" xref=83.63 yref=22.01'+ \
                          ' migramax=2.0 migrabins=10'+ \
                          ' logfile="ctedispcube_cmd2.log" debug=yes chatter=2'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctedispcube --help
        self._check_help(ctedispcube)

        # Return
        return

    # Test ctedispcube from Python
    def _test_python(self):
        """
        Test ctedispcube from Python
        """
        # Allocate ctedispcube
        edispcube = ctools.ctedispcube()

        # Check that empty ctedispcube tool holds an energy dispersion
        # cube that has no energy and migration bins
        self._check_cube(edispcube.edispcube(), nenergies=0, nmigras=0)

        # Check that saving does not nothing
        edispcube['outcube'] = 'ctedispcube_py0.fits'
        edispcube['logfile'] = 'ctedispcube_py0.log'
        edispcube.logFileOpen()
        edispcube.save()
        self.test_assert(not os.path.isfile('ctedispcube_py0.fits'),
             'Check that no energy dispersion cube has been created')

        # Check that clearing does not lead to an exception or segfault
        edispcube.clear()

        # Now set ctedispcube parameters
        edispcube['inobs']     = self._events
        edispcube['incube']    = 'NONE'
        edispcube['caldb']     = self._caldb
        edispcube['irf']       = self._irf
        edispcube['ebinalg']   = 'LOG'
        edispcube['emin']      = 1.0
        edispcube['emax']      = 100.0
        edispcube['enumbins']  = 5
        edispcube['nxpix']     = 10
        edispcube['nypix']     = 10
        edispcube['binsz']     = 0.4
        edispcube['coordsys']  = 'CEL'
        edispcube['proj']      = 'CAR'
        edispcube['xref']      = 83.63
        edispcube['yref']      = 22.01
        edispcube['migramax']  = 2.0
        edispcube['migrabins'] = 10
        edispcube['outcube']   = 'ctedispcube_py1.fits'
        edispcube['logfile']   = 'ctedispcube_py1.log'
        edispcube['chatter']   = 2

        # Run ctedispcube tool
        edispcube.logFileOpen()  # Make sure we get a log file
        edispcube.run()
        edispcube.save()

        # Check result file
        self._check_result_file('ctedispcube_py1.fits')

        # Copy ctedispcube tool
        cpy_edispcube = edispcube.copy()

        # Check energy dispersion cube of ctedispcube copy
        self._check_cube(cpy_edispcube.edispcube())

        # Execute copy of ctedispcube tool again, now with a higher chatter
        # level than before
        cpy_edispcube['emin']      = 0.2
        cpy_edispcube['emax']      = 150.0
        cpy_edispcube['outcube']   = 'ctedispcube_py2.fits'
        cpy_edispcube['logfile']   = 'ctedispcube_py2.log'
        cpy_edispcube['addbounds'] = True
        cpy_edispcube['chatter']   = 3
        cpy_edispcube.logFileOpen()  # Needed to get a new log file
        cpy_edispcube.execute()

        # Check result file
        self._check_result_file('ctedispcube_py2.fits', nenergies=8)

        # Now clear copy of ctedispcube tool
        cpy_edispcube.clear()

        # Check that the cleared copy has also cleared the energy dispersion
        # cube
        self._check_cube(cpy_edispcube.edispcube(), nenergies=0, nmigras=0)

        # Get mixed observation container
        obs = self._obs_mixed()

        # Set-up ctedispcube from observation container
        edispcube = ctools.ctedispcube(obs)
        edispcube['incube']    = 'NONE'
        edispcube['caldb']     = self._caldb
        edispcube['irf']       = self._irf
        edispcube['ebinalg']   = 'POW'
        edispcube['emin']      = 1.0
        edispcube['emax']      = 100.0
        edispcube['enumbins']  = 5
        edispcube['ebingamma'] = 1.0
        edispcube['nxpix']     = 10
        edispcube['nypix']     = 10
        edispcube['binsz']     = 0.4
        edispcube['coordsys']  = 'CEL'
        edispcube['proj']      = 'CAR'
        edispcube['xref']      = 83.63
        edispcube['yref']      = 22.01
        edispcube['migramax']  = 2.0
        edispcube['migrabins'] = 10
        edispcube['addbounds'] = True
        edispcube['outcube']   = 'ctedispcube_py3.fits'
        edispcube['logfile']   = 'ctedispcube_py3.log'
        edispcube['chatter']   = 4

        # Execute ctedispcube tool
        edispcube.logFileOpen()  # Make sure we get a log file
        edispcube.execute()

        # Check result file
        self._check_result_file('ctedispcube_py3.fits')

        # Set-up ctexpcube from event list and counts cube
        edispcube = ctools.ctedispcube()
        edispcube['inobs']     = self._events
        edispcube['incube']    = self._cntcube
        edispcube['caldb']     = self._caldb
        edispcube['irf']       = self._irf
        edispcube['migramax']  = 2.0
        edispcube['migrabins'] = 10
        edispcube['outcube']   = 'ctedispcube_py4.fits'
        edispcube['logfile']   = 'ctedispcube_py4.log'
        edispcube['chatter']   = 2

        # Run ctedispcube tool
        edispcube.logFileOpen()  # Make sure we get a log file
        edispcube.execute()

        # Check result file
        self._check_result_file('ctedispcube_py4.fits')

        # Return
        return

    # Check result file
    def _check_result_file(self, filename, nenergies=6):
        """
        Check result file

        Parameters
        ----------
        filename : str
            Energy dispersion cube file name
        nenergies : int, optional
            Number of energy bins
        """
        # Open energy dispersion cube
        cube = gammalib.GCTACubeEdisp(filename)

        # Check cube
        self._check_cube(cube, nenergies=nenergies)

        # Return
        return

    # Check energy dispersion cube
    def _check_cube(self, cube, nenergies=6, nmigras=10):
        """
        Check energy dispersion cube

        Parameters
        ----------
        cube : `~gammalib.GCTACubeEdisp`
            Energy dispersion cube
        nenergies : int, optional
            Number of energy bins
        nmigras : int, optional
            Number of migration bins
        """
        # Check dimensions
        self.test_value(len(cube.energies()), nenergies,
             'Check number of energy maps')
        self.test_value(len(cube.migras()), nmigras,
             'Check number of migration bins')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csresspec script.
#
# Copyright (C) 2017-2020 Luigi Tibaldo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# =============================== #
# Test class for csresspec script #
# =============================== #
class Test(test):
    """
    Test class for csresspec script

    This test class makes unit tests for the csresspec script by using it
    from the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set members
        self._inonoff     = self._datadir + '/onoff_obs.xml'
        self._onoff_model = self._datadir + '/onoff_model.xml'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csresspec')

        # Append tests
        self.append(self._test_cmd, 'Test csresspec on command line')
        self.append(self._test_python, 'Test csresspec from Python')

        # Return
        return

    # Test csresspec on command line
    def _test_cmd(self):
        """
        Test csresspec on the command line.
        """
        # Set script name
        csresspec = self._script('csresspec')

        # Setup csresspec command
        cmd = csresspec + ' inobs="' + self._events + '"' + \
                          ' outfile="csresspec_cmd1.fits"' + \
                          ' inmodel="' + self._model + '"' + \
                          ' caldb="' + self._caldb + '" irf="' + self._irf + '"' + \
                          ' emin=1.0 emax=100.0 enumbins=2 ebinalg="LOG"' + \
                          ' mask=no algorithm="SUB"' + \
                          ' logfile="csresspec_cmd1.log" chatter=4'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
                         'Check successful execution from command line')

        # Check result file
        self._check_result_file('csresspec_cmd1.fits', 2, 5)

        # Setup csresspec command
        cmd = csresspec + ' inobs="event_file_that_does_not_exist.fits"' + \
                          ' outfile="csresspec_cmd2.fits"' + \
                          ' inmodel="' + self._model + '"' + \
                          ' caldb="' + self._caldb + '" irf="' + self._irf + '"' + \
                          ' emin=1.0 emax=100.0 enumbins=2 ebinalg="LOG"' + \
                          ' mask=no algorithm="SUBDIVSQRT"' + \
                          ' logfile="csresspec_cmd2.log" debug=yes chatter=2'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csresspec --help
        self._check_help(csresspec)

        # Return
        return

    # Test csresspec from Python
    def _test_python(self):
        """
        Test csresspec from Python
        """
        # Set-up csresspec for event list
        resspec = cscripts.csresspec()
        resspec['inobs']     = self._events
        resspec['outfile']   = 'csresspec_py1.fits'
        resspec['inmodel']   = self._model
        resspec['caldb']     = self._caldb
        resspec['irf']       = self._irf
        resspec['ebinalg']   = 'LOG'
        resspec['emin']      = 1.0
        resspec['emax']      = 100.0
        resspec['enumbins']  = 2
        resspec['mask']      = False
        resspec['algorithm'] = 'SUB'
        resspec['logfile']   = 'csresspec_py1.log'
        resspec['chatter']   = 2

        # Run csresspec script
        resspec.logFileOpen()  # Make sure we get a log file
        resspec.run()
        resspec.save()

        # Check result file
        self._check_result_file('csresspec_py1.fits', 2, 5)

        # Set-up csresspec for event list with computation of individual
        # model components
        resspec = cscripts.csresspec()
        resspec['inobs']      = self._events
        resspec['outfile']    = 'csresspec_py2.fits'
        resspec['inmodel']    = self._model
        resspec['caldb']      = self._caldb
        resspec['irf']        = self._irf
        resspec['ebinalg']   = 'POW'
        resspec['emin']       = 1.0
        resspec['emax']       = 100.0
        resspec['enumbins']   = 2
        resspec['ebingamma']  = 1.0
        resspec['mask']       = False
        resspec['algorithm']  = 'SUBDIV'
        resspec['components'] = True
        resspec['logfile']    = 'csresspec_py2.log'
        resspec['chatter']    = 2

        # Run csresspec script
        resspec.execute()

        # Check result file
        self._check_result_file('csresspec_py2.fits', 2, 7)

        # Set-up csresspec for counts cube
        resspec = cscripts.csresspec()
        resspec['inobs']     = self._cntcube
        resspec['modcube']   = 'NONE'
        resspec['expcube']   = 'NONE'
        resspec['psfcube']   = 'NONE'
        resspec['edispcube'] = 'NONE'
        resspec['bkgcube']   = 'NONE'
        resspec['caldb']     = self._caldb
        resspec['irf']       = self._irf
        resspec['inmodel']   = self._model
        resspec['outfile']   = 'csresspec_py3.fits'
        resspec['mask']      = False
        resspec['algorithm'] = 'SUBDIVSQRT'
        resspec['logfile']   = 'csresspec_py3.log'
        resspec['chatter']   = 3

        # Run csresspec script
        resspec.execute()

        # Check result file
        self._check_result_file('csresspec_py3.fits', 5, 5)

        # Set-up csresspec for counts cube with source mask
        resspec = cscripts.csresspec()
        resspec['inobs']     = self._cntcube
        resspec['modcube']   = 'NONE'
        resspec['expcube']   = 'NONE'
        resspec['psfcube']   = 'NONE'
        resspec['edispcube'] = 'NONE'
        resspec['bkgcube']   = 'NONE'
        resspec['caldb']     = self._caldb
        resspec['irf']       = self._irf
        resspec['inmodel']   = self._model
        resspec['outfile']   = 'csresspec_py4.fits'
        resspec['mask']      = True
        resspec['ra']        = 83.63
        resspec['dec']       = 22.01
        resspec['rad']       = 0.2
        resspec['regfile']   = 'NONE'
        resspec['algorithm'] = 'SUBDIVSQRT'
        resspec['logfile']   = 'csresspec_py4.log'
        resspec['chatter']   = 3

        # Run csresspec script
        resspec.execute()

        # Check result file
        self._check_result_file('csresspec_py4.fits', 5, 5)

        # Set-up csresspec for On/Off observations
        resspec = cscripts.csresspec()
        resspec['inobs']     = self._inonoff
        resspec['inmodel']   = self._onoff_model
        resspec['outfile']   = 'csresspec_py5.fits'
        resspec['algorithm'] = 'SIGNIFICANCE'
        resspec['stack']     = False
        resspec['logfile']   = 'csresspec_py5.log'
        resspec['chatter']   = 3

        # Run csresspec script
        resspec.execute()

        # Check result file
        self._check_result_file('csresspec_py5.fits', 30, 8, nobs=2)

        # Set-up csresspec for On/Off observations
        # in stacked mode and with WSTAT background
        resspec = cscripts.csresspec()
        resspec['inobs']     = self._inonoff
        resspec['inmodel']   = self._onoff_model
        resspec['outfile']   = 'csresspec_py6.fits'
        resspec['algorithm'] = 'SIGNIFICANCE'
        resspec['statistic'] = 'WSTAT'
        resspec['stack']     = True
        resspec['logfile']   = 'csresspec_py6.log'
        resspec['chatter']   = 3

        # Run csresspec script
        resspec.execute()

        # Check result file
        self._check_result_file('csresspec_py6.fits', 30, 8)

        # Return
        return

    # Check result file
    def _check_result_file(self, filename, enumbins, ncols, nobs=1):
        """
        Check result file
        """
        # Load residual file
        fits = gammalib.GFits(filename)

        # Check number of HDUs = observations (+ first empty image)
        self.test_value(fits.size(), 1 + nobs, 'Number of observations')

        # Check first hdu with residuals
        table = fits[1]

        # Check number of rows = number of energy bins
        self.test_value(table.nrows(), enumbins,
                        'Check for %d rows in residuals' % enumbins)

        # Check number of columns, 2 for energy bounds + 3 for 3D or 6 for
        # On/Off + 1 for each model component if components = yes
        self.test_value(table.ncols(), ncols,
                        'Check for %d columns in spectrum' % ncols)

        # Return
        return



#!/usr/bin/env python
# ==========================================================================
# Shows the content of a model container
#
# Copyright (C) 2015-2018 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools
from cscripts import modutils


# ================= #
# csmodelinfo class #
# ================= #
class csmodelinfo(ctools.cscript):
    """
    Shows the content of a model container

    The script is useful to show, e.g., the number of sources in a model
    container, the background models, free and fixed parameters etc.
    The script can also dump the models into a region file to for display
    with DS9.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Initialise class members
        self._models = gammalib.GModels()

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation
        """
        # Get models
        self._models = gammalib.GModels(self['inmodel'].filename())
        
        # Query hidden parameters for region file
        self['pnt_type'].string()
        self['pnt_mark_size'].integer()
        self['show_labels'].boolean()
        self['width'].integer()
        self['fontfamily'].string()
        self['fontsize'].integer()
        self['fontweight'].string()
        self['fontslant'].string()
        self['show_ext_type'].boolean()
        self['free_color'].string()
        self['fixed_color'].string()

        # Query ahead DS9 filename
        if self._read_ahead():
            self['outds9file'].query()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)
        
        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()
        
        # Optionally log models
        if self._logVerbose():
            self._log('\n')
            self._log.header1('Models')
            self._log('\n')
            self._log(str(self._models))
            self._log('\n')
        
        # Initialise output values 
        types           = {}
        instruments     = {}
        ts              = {}
        pars_at_limit   = {}
        free_src_pars   = {}
        n_par_total     = 0
        n_par_free      = 0
        n_par_fixed     = 0
        n_par_at_limit  = 0
        n_par_free_bkg  = 0
        n_par_free_src  = 0
        n_par_free_spec = 0
        n_par_free_spat = 0
        n_par_free_temp = 0
        
        # Loop over models and initialise dictionaries with model names
        for model in self._models:
            types[model.type()]              = 0
            instruments[model.instruments()] = 0
            pars_at_limit[model.name()]      = []

        # Loop over models again and fill values to be dumped
        for model in self._models:
            
            # Model types
            types[model.type()] += 1
            
            # Model instruments
            instruments[model.instruments()] += 1 
            
            # TS
            if model.has_ts():
                ts[model.name()] = model.ts()
                
            # Skymodel?
            skymodel = (model.classname() == 'GModelSky')
            if skymodel:
                free_src_pars[model.name()] = []                
            
            # Loop over model parameters
            for par in model:
                
                # Increment total parameters
                n_par_total += 1
                
                # Check if parameter is at limit
                if par.has_range():
                    if par.value() == par.min() or par.value() == par.max():
                        pars_at_limit[model.name()].append(par.name())   
                        n_par_at_limit += 1 
                        
                # Check if model parameter is free/fixed and
                # increment respective variable      
                if par.is_free():         
                    n_par_free += 1
                    if skymodel:
                        free_src_pars[model.name()].append(par)
                        n_par_free_src += 1
                    else:
                        n_par_free_bkg += 1
                else:
                    n_par_fixed += 1
            
            # If sky model, divide free parameters into spatial, spectral
            # and temporal
            if skymodel:
                for par in model.spectral():
                    if par.is_free():
                        n_par_free_spec += 1
                for par in model.spatial():
                    if par.is_free():
                        n_par_free_spat += 1
                for par in model.temporal():
                    if par.is_free():
                        n_par_free_temp += 1     

        # Optionally log model information
        if self._logTerse():
        
            # Log summary
            self._log_header1(gammalib.TERSE, 'Summary')
        
            # Log instruments
            self._log.header3('Instrument specific models')
            for inst, n_inst in instruments.items():
                if inst == '':
                    inst = 'All'
                self._log_value(gammalib.TERSE, inst, n_inst)
        
            # Log model types
            self._log_header3(gammalib.TERSE, 'Model types')
            for modeltype, n_types in types.items():
                self._log_value(gammalib.TERSE, modeltype, n_types)
        
            # Log model parameter information
            self._log_header1(gammalib.TERSE,'Parameter information')
            self._log_value(gammalib.TERSE,'All parameters', n_par_total)
            self._log_value(gammalib.TERSE,'Fixed parameters', n_par_fixed)
            self._log_value(gammalib.TERSE,'Free parameters (total)', n_par_free)
            self._log_value(gammalib.TERSE,'Free background parameters', n_par_free_bkg)
            self._log_value(gammalib.TERSE,'Free source parameters', n_par_free_src)
            if n_par_free_spec > 0:
                self._log_value(gammalib.TERSE,'Free spectral parameters', n_par_free_spec)
            if n_par_free_spat > 0:
                self._log_value(gammalib.TERSE,'Free spatial parameters', n_par_free_spat)
            if n_par_free_temp > 0:
                self._log_value(gammalib.TERSE,'Free temporal parameters', n_par_free_temp)
            self._log_value(gammalib.TERSE,'Parameters at limit', n_par_at_limit)
        
            # Log parameters at limit (if any)
            if n_par_at_limit > 0:
                self._log_header3(gammalib.TERSE, 'Parameters at limit')
                for source, parameter in pars_at_limit.items():   
                    if len(parameter):  
                        for par in parameter:
                            self._log_value(gammalib.TERSE, source, par)
        
            # Optionally log free parameters
            if len(free_src_pars):
                self._log_header2(gammalib.EXPLICIT, 'Free source parameters')
                for source, parameter in free_src_pars.items():   
                    if len(parameter): 
                        self._log.header3(source)
                        for par in parameter:
                            self._log_string(gammalib.EXPLICIT, str(par)+'\n')
        
            # Log TS values if available
            if len(ts):       
                self._log_header3(gammalib.TERSE, 'Test statistics')
                for source, tsvalue in ts.items():
                    self._log_value(gammalib.TERSE, source, tsvalue)
                
        # Return
        return

    def save(self):
        """ 
        Save models to ds9 region file
        """
        # Check if DS9 file is valid
        if self['outds9file'].is_valid():

            # Get output filename
            ds9file = self['outds9file'].filename()

            # Write header
            self._log_header1(gammalib.TERSE, 'Save models in DS9 file')
            
            # Log filename
            self._log_value(gammalib.NORMAL, 'DS9 filename', ds9file.url())

            # Get ds9 parameters
            pnt_type      = self['pnt_type'].string()
            pnt_mark_size = self['pnt_mark_size'].integer()
            show_labels   = self['show_labels'].boolean()
            free_color    = self['free_color'].string()
            fixed_color   = self['fixed_color'].string()
            width         = self['width'].integer()
            fontfamily    = self['fontfamily'].string()
            fontsize      = self['fontsize'].integer()
            fontweight    = self['fontweight'].string()
            fontslant     = self['fontslant'].string()

            # Save DS9 region file
            errors = modutils.models2ds9file(self._models, ds9file.url(),
                                             pnt_type=pnt_type,
                                             pnt_mark_size=pnt_mark_size,
                                             show_labels=show_labels,
                                             free_color=free_color,
                                             fixed_color=fixed_color,
                                             width=width,
                                             fontfamily=fontfamily,
                                             fontsize=fontsize,
                                             fontweight=fontweight,
                                             fontslant=fontslant)

            # Log errors
            if len(errors):
                self._log(errors)
    
        # Return
        return 


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csmodelinfo(sys.argv)
    
    # Execute application
    app.execute()



#!/usr/bin/env python
# ==========================================================================
# Find observations from an IACT data store
#
# Copyright (C) 2016-2018 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import os
import json
import gammalib
import ctools


# =============== #
# csfindobs class #
# =============== #
class csfindobs(ctools.cscript):
    """
    Find observations from an IACT data store
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Set name
        self._datapath     = os.getenv('VHEFITS','')
        self._prodname     = ''
        self._select_radec = True
        self._radius       = 0.0
        self._ra           = 0.0
        self._dec          = 0.0
        self._obs_index    = ''
        self._runs         = []

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation.
        """
        # Query datapath. If the parameter is not NONE then use it, otherwise
        # use the datapath from the VHEFITS environment variable
        datapath = self['datapath'].string()
        if gammalib.toupper(datapath) != 'NONE':
            self._datapath = datapath
        else:
            self._datapath = os.getenv('VHEFITS','')

        # Expand environment
        self._datapath = gammalib.expand_env(self._datapath)

        # Get production name
        self._prodname = self['prodname'].string()

        # Master index file name
        master_indx = self['master_indx'].string()

        # Initialise flag if spatial selection is required
        self._select_radec = True

        # Initialise invalid radius
        self._radius = 0.0

        # Check for validity of spatial parameters
        if (self['ra'].is_valid() and
            self['dec'].is_valid() and 
            self['rad'].is_valid()):
            
            # Read spatial parameters
            self._ra     = self['ra'].real()
            self._dec    = self['dec'].real()
            self._radius = self['rad'].real()

        # ... otherwise signal that there are no spatial parameters for
        # selection
        else:
            self._select_radec = False
        
        # Check Radius for validity
        if self._radius <= 0.0:
            self._select_radec = False
            
        # Query other parameters
        self['min_qual'].integer()
        self['expression'].string()

        # Read ahead output parameters
        if self._read_ahead():
            self['outfile'].filename()
        
        # Set filename of JSON master file and raise an exception if the file
        # does not exist
        master_file = os.path.join(self._datapath, master_indx)
        if not os.path.isfile(master_file):
            msg = ('FITS data store not available. No master index file found '
                   'at "%s". Make sure the file is copied from the server and '
                   'your datapath is set correctly.' % master_file)
            raise RuntimeError(msg)

        # Open and load JSON master file. If the "dataset" key is not available
        # then raise an exception
        json_data = open(master_file).read()
        data      = json.loads(json_data)
        if not 'datasets' in data:
            msg = ('Key "datasets" not available in master index file.')
            raise RuntimeError(msg)

        # Get configurations from JSON master file
        configs = data['datasets']

        # Initialise obs index file
        self._obs_index = ''

        # Get name of observation index file
        for config in configs:
            if self._prodname == config['name']:
                self._obs_index = str(os.path.join(self._datapath,
                                                   config['obsindx']))
                break

        # If the observation index file name is empty then raise an exception
        if self._obs_index == '':
            msg = ('FITS data store "%s" not available. Run csiactdata to get '
                   'a list of available storage names.' % self._prodname)
            raise RuntimeError(msg)

        # If the observation index file is not a FITS file then raise an
        # exception
        filename = gammalib.GFilename(self._obs_index+'[OBS_INDEX]')
        if not filename.is_fits():
            msg = ('Observation index file "%s[OBS_INDEX]" for FITS data store '
                   '"%s" not available. Check your master index file or run '
                   'csiactdata to get a list of available storage names.' %
                   (self._obs_index, self._prodname))
            raise RuntimeError(msg)

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write header into logger
        self._log_header1(gammalib.TERSE, 'Find observations')
        
        # Initialise run list
        self._runs = []
        
        # Initialise selection expression
        expr = ''
        
        # If spatial selection is requested then add an angular separation
        # expression to the expression string
        if self._select_radec:
            expr += 'ANGSEP('+str(self._ra)+','+str(self._dec)+ \
                    ',RA_PNT,DEC_PNT)<='+str(self._radius)
        
        # Add '&&' connector if expression is not empty
        if len(expr):
            expr += '&&'
        
        # Add always quality expression
        expr += 'QUALITY<='+str(self['min_qual'].integer())
        
        # Add user expression is one has been specified
        expression = self['expression'].string()    
        if (expression != 'NONE' and expression != 'INDEF' and
            len(expression) > 0):
            expr += '&&' + expression   

        # Write the expression into the logger
        self._log_value(gammalib.NORMAL, 'Expression', expr)

        # Set filename including the selection expression
        filename = self._obs_index+'[OBS_INDEX]['+expr+']'
        
        # Open observation index FITS file
        fits = gammalib.GFits(filename)

        # If there are entries in the observation index table then append the
        # observations identifiers to the run list
        obs_index = fits['OBS_INDEX']
        if obs_index.nrows() > 0:
            for i in range(obs_index.nrows()):
                self._runs.append(obs_index['OBS_ID'][i])

        # Write the number of observations into the logger
        self._log_value(gammalib.TERSE, 'Observations', len(self._runs))

        # Write the observation identifiers into the logger
        if len(self._runs) > 0:
            for i, run in enumerate(self._runs):
                self._log_value(gammalib.NORMAL, 'Observation %d' % (i+1), run)

        # Close FITS file
        fits.close()

        # Return
        return
    
    def save(self):
        """
        Save runlist
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save runlist')

        # Get filename
        outfile = self['outfile'].filename()

        # If the runlist file exists but the clobber flag is not set then signal
        # that runlist file exists already
        if outfile.exists() and not self._clobber():
            msg = ('File "'+outfile+'" exists already, runlist not saved. '
                   'Set clobber=yes to overwrite the file.\n')
            self._log_string(gammalib.TERSE, msg)

        # ... otherwise save the runlist into a file
        else:

            # Log file name
            self._log_value(gammalib.NORMAL, 'Runlist file', outfile.url())

            # Write all runs to file
            f = open(outfile.url(),'w')
            for run in self._runs:
                f.write(str(run)+' \n')
            f.close()

        # Return
        return

    def runs(self):
        """ 
        Return run list

        Returns
        -------
        runs : list of str
            A list of runs
        """
        # Return
        return self._runs
    

# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csfindobs(sys.argv)
    
    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctphase tool.
#
# Copyright (C) 2017-2018 by Joshua Cardenzana
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from testing import test


# ============================ #
# Test class for ctselect tool #
# ============================ #
class Test(test):
    """
    Test class for ctselect tool
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set test data
        self._model_file     = self._datadir + '/model_temporal_phasecurve.xml'
        self._invalid_events = self._datadir + '/invalid_event_list.fits'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctphase')

        # Append tests
        self.append(self._test_cmd, 'Test ctphase on command line')
        self.append(self._test_python, 'Test ctphase from Python')

        # Return
        return

    # Test ctphase on command line
    def _test_cmd(self):
        """
        Test ctphase on the command line
        """
        # Set tool name
        ctphase = self._tool('ctphase')

        # Setup ctphase command
        cmd = ctphase+' inobs="'+self._events+'"'+ \
                      ' outobs="ctphase_cmd1.fits"'+ \
                      ' inmodel="'+self._model_file+'" srcname="Crab"'+ \
                      ' logfile="ctphase_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('ctphase_cmd1.fits')

        # Setup ctphase command
        cmd = ctphase+' inobs="event_file_that_does_not_exist.fits"'+ \
                      ' outobs="ctphase_cmd2.fits"'+ \
                      ' inmodel="'+self._model_file+'" srcname="Crab"'+ \
                      ' logfile="ctphase_cmd2.log" debug=yes chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctphase --help
        self._check_help(ctphase)

        # Return
        return

    # Test ctphase from Python
    def _test_python(self):
        """
        Test ctphase from Python
        """
        # Allocate empty ctphase tool
        phase = ctools.ctphase()

        # Check that empty ctphase tool holds an empty observation
        self._check_obs(phase.obs(), nobs=0)

        # Check that saving does nothing
        phase['inmodel'] = self._model_file
        phase['srcname'] = 'Crab'
        phase['outobs']  = 'ctphase_py0.fits'
        phase['logfile'] = 'ctphase_py0.log'
        phase.logFileOpen()
        phase.save()
        self.test_assert(not os.path.isfile('ctphase_py0.fits'),
             'Check that no event list has been created')

        # Check that clearing does not lead to an exception or segfault
        phase.clear()

        # Now set ctphase parameters
        phase['inobs']   = self._events
        phase['inmodel'] = self._model_file
        phase['srcname'] = 'Crab'
        phase['outobs']  = 'ctphase_py1.fits'
        phase['logfile'] = 'ctphase_py1.log'
        phase['chatter'] = 2

        # Run ctphase tool
        phase.logFileOpen()   # Make sure we get a log file
        phase.run()
        phase.save()

        # Check result file
        self._check_result_file('ctphase_py1.fits')

        # Clear phase tool
        phase.clear()

        # Now run ctphase tool without a model file
        phase['inobs']   = self._events
        phase['inmodel'] = 'NONE'
        phase['srcname'] = 'NONE'
        phase['mjd']     = 58849.0
        phase['phase']   = 0.0
        phase['f0']      = 1.0
        phase['f1']      = 0.1
        phase['f2']      = 0.01
        phase['outobs']  = 'ctphase_py2.fits'
        phase['logfile'] = 'ctphase_py2.log'
        phase['chatter'] = 3

        # Execute ctphase tool
        phase.logFileOpen()   # Make sure we get a log file
        phase.execute()

        # Check result file
        self._check_result_file('ctphase_py2.fits')

        # Copy ctphase tool
        cpy_phase = phase.copy()

        # Clear phase tool
        phase.clear()

        # Execute copy
        cpy_phase['outobs']  = 'ctphase_py3.fits'
        cpy_phase['logfile'] = 'ctphase_py3.log'
        cpy_phase['chatter'] = 4
        cpy_phase['publish'] = True
        cpy_phase.logFileOpen()  # Needed to get a new log file
        cpy_phase.execute()

        # Check result file
        self._check_result_file('ctphase_py3.fits')

        # Now run ctphase tool without a model file and a significant
        # MJD offset
        phase['inobs']   = self._events
        phase['inmodel'] = 'NONE'
        phase['srcname'] = 'NONE'
        phase['mjd']     = 58860.0
        phase['phase']   = 0.0
        phase['f0']      = 1.0
        phase['f1']      = 0.1
        phase['f2']      = 0.0001
        phase['outobs']  = 'ctphase_py4.fits'
        phase['logfile'] = 'ctphase_py4.log'
        phase['chatter'] = 3

        # Execute ctphase tool
        phase.logFileOpen()   # Make sure we get a log file
        phase.execute()

        # Check result file
        self._check_result_file('ctphase_py4.fits')

        # Test invalid source name
        self.test_try('Test ctphase with invalid source name')
        try:
            phase = ctools.ctphase()
            phase['inobs']   = self._events
            phase['inmodel'] = self._model_file
            phase['srcname'] = 'Venus'
            phase['outobs']  = 'ctphase_py5.fits'
            phase['logfile'] = 'ctphase_py5.log'
            phase.logFileOpen()
            phase.execute()
            self.test_try_failure('Exception not thrown')
        except (ValueError):
            self.test_try_success()

        # Test invalid model type
        self.test_try('Test ctphase with invalid model')
        try:
            phase = ctools.ctphase()
            phase['inobs']   = self._events
            phase['inmodel'] = self._model
            phase['srcname'] = 'Background'
            phase['outobs']  = 'ctphase_py6.fits'
            phase['logfile'] = 'ctphase_py6.log'
            phase.logFileOpen()
            phase.execute()
            self.test_try_failure('Exception not thrown')
        except (ValueError):
            self.test_try_success()

        # Test invalid model without temporal phase curve model
        self.test_try('Test ctphase with invalid model')
        try:
            phase = ctools.ctphase()
            phase['inobs']   = self._events
            phase['inmodel'] = self._model
            phase['srcname'] = 'Crab'
            phase['outobs']  = 'ctphase_py7.fits'
            phase['logfile'] = 'ctphase_py7.log'
            phase.logFileOpen()
            phase.execute()
            self.test_try_failure('Exception not thrown')
        except (ValueError):
            self.test_try_success()

        # Return
        return

    # Check result file
    def _check_result_file(self, filename, nevents=253):
        """
        Check result file

        Parameters
        ----------
        filename : str
            Event list file name
        nevents : int, optional
            Expected number of events
        """
        # Open event list
        events = gammalib.GCTAEventList(filename)

        # Check event list
        self.test_value(events.size(), nevents, 'Check number of events')

        # Check phases
        phases      = [events[i].phase() for i in range(events.size())]
        test_result = int(sum(phases) > 0)
        self.test_value(test_result, 1, 'Check whether phase information is set')

        # Return
        return

    # Check observation and event list
    def _check_obs(self, obs, nobs=1):
        """
        Check observation and event list

        Parameters
        ----------
        obs : `~gammalib.GObservations`
            Models
        nobs : int, optional
            Expected number of observations
        """
        # Check number of observations
        self.test_value(obs.size(), nobs, 'Check number of observations')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csobs2caldb script.
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import cscripts
from testing import test


# ================================= #
# Test class for csobs2caldb script #
# ================================= #
class Test(test):
    """
    Test class for csobs2caldb script

    This test class makes unit tests for the csobs2caldb script by using it
    from the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set members
        self._inobs = self._datadir + '/obs_unbinned.xml'

        # Clean calibration database
        os.system('rm -rf caldb >/dev/null 2>&1')
        os.system('mkdir -p caldb >/dev/null 2>&1')

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csobs2caldb')

        # Append tests
        self.append(self._test_cmd, 'Test csobs2caldb on command line')
        self.append(self._test_python, 'Test csobs2caldb from Python')

        # Return
        return

    # Test csobs2caldb on command line
    def _test_cmd(self):
        """
        Test csobs2caldb on the command line
        """
        # Set script name
        csobs2caldb = self._script('csobs2caldb')

        # Setup csobs2caldb command
        cmd = csobs2caldb+' inobs="'+self._inobs+'"'+ \
                          ' caldb="cta" irf="Cmd1" rootdir="caldb"'+ \
                          ' logfile="csobs2caldb_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check response
        self._check_response('cta', 'Cmd1')

        # Setup csobs2caldb command
        cmd = csobs2caldb+' inobs="xml_file_that_does_not_exist.xml"'+ \
                          ' caldb="cta" irf="Cmd2" rootdir="caldb"'+ \
                          ' logfile="csobs2caldb_cmd2.log" debug=yes'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csobs2caldb --help
        self._check_help(csobs2caldb)

        # Return
        return

    # Test csobs2caldb from Python
    def _test_python(self):
        """
        Test csobs2caldb from Python
        """
        # Set-up csobs2caldb using "inobs" parameter
        irf = cscripts.csobs2caldb()
        irf['inobs']   = self._inobs
        irf['caldb']   = 'cta'
        irf['irf']     = 'Py1'
        irf['rootdir'] = 'caldb'
        irf['logfile'] = 'csobs2caldb_py1.log'
        irf['chatter'] = 2

        # Run csobs2caldb script and save IRF
        irf.logFileOpen()   # Make sure we get a log file
        irf.run()
        irf.save()

        # Check response
        self._check_response('cta', 'Py1')

        # Load observation container
        obs = gammalib.GObservations(self._inobs)

        # Set-up csobs2caldb using observation container
        irf = cscripts.csobs2caldb(obs)
        irf['caldb']   = 'cta'
        irf['irf']     = 'Py2'
        irf['rootdir'] = 'caldb'
        irf['logfile'] = 'csobs2caldb_py2.log'
        irf['chatter'] = 3

        # Execute csobs2caldb script
        irf.execute()

        # Check response
        self._check_response('cta', 'Py2')

        # Return
        return

    # Check response
    def _check_response(self, caldb, irf):
        """
        Check response
        """
        # Open calibration database
        db = gammalib.GCaldb('caldb')
        db.open('cta', caldb)

        # Get filenames of response components
        expr  = 'NAME('+irf+')'
        aeff  = gammalib.GFilename(db.filename('','','EFF_AREA','','',expr))
        psf   = gammalib.GFilename(db.filename('','','RPSF','','',expr))
        edisp = gammalib.GFilename(db.filename('','','EDISP','','',expr))
        bkg   = gammalib.GFilename(db.filename('','','BKG','','',expr))

        # Check whether files exist
        self.test_assert(aeff.exists(), 'Effective area file exists')
        self.test_assert(psf.exists(), 'Point spread function file exists')
        self.test_assert(edisp.exists(), 'Energy dispersion file exists')
        self.test_assert(bkg.exists(), 'Background file exists')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Display TS distribution generated by cstsdist
#
# Copyright (C) 2011-2017 Jurgen Knodlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import csv
import math
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()
try:
    import numpy as np
except ImportError:
    print('This script needs the "numpy" module')
    sys.exit()


# ==================== #
# Read TS distribution #
# ==================== #
def read_ts(filename, tsname='TS'):
    """
    Read TS distribution values from CSV file

    Parameters
    ----------
    filename : str
        Name of CSV file
    tsname : str, optional
        Column name of TS

    Returns
    -------
    a : `numpy.array`
        Array of TS values
    """
    # Initialise list
    values = []

    # Open reader
    reader = csv.reader(open(filename, 'r'), delimiter=',')

    # Read rows
    first = True
    index = -1
    for row in reader:

        # Get column index if first row
        if first:
            try:
                index = row.index(tsname)
            except:
                print('ERROR: Column "'+tsname+'" not found in file')
                print(row)
                raise NameError(tsname)

        # Handle data rows
        else:
            values.append(float(row[index]))

        # Flag that first row has been passed
        first = False

    # Create numpy array
    a = np.array(values)

    # Return array
    return a


# =========== #
# Compute erf #
# =========== #
def erf(x):
    """
    Compute error function

    Parameters
    ----------
    x : float
        Value

    Returns
    -------
    erf(-x) : float
        Error function
    """
    # save the sign of x
    sign = 1
    if x < 0: 
        sign = -1
    x = abs(x)

    # constants
    a1 =  0.254829592
    a2 = -0.284496736
    a3 =  1.421413741
    a4 = -1.453152027
    a5 =  1.061405429
    p  =  0.3275911

    # A&S formula 7.1.26
    t = 1.0/(1.0 + p*x)
    y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*math.exp(-x*x)

    # Return result
    return sign*y # erf(-x) = -erf(x)


# =============================== #
# Display cumulative distribution #
# =============================== #
def dist_cdf(values, nbins, title, plotfile, expect=''):
    """
    Display cumulative TS distribution and overplots expectation

    Parameters
    ----------
    values : float
        TS values
    nbins : int
        Number of bins
    title : str
        Plot title
    plotfile : str
        Plot file name
    expect : str, optional
        If 'show', show expectation
    """
    # Set range and adapt number of bins. We make sure that we have a
    # histogram bin centred on 0 that will capture the TS=0 part.
    max_value = max(values)
    binsize   = max_value/nbins
    min_value = 0.5*binsize
    while min_value > min(values):
        min_value -= binsize
        nbins     += 1

    # Create histogram
    hist_x      = []
    hist_y      = []
    hist_expect = []
    norm        = 0.5*len(values)
    for i in range(nbins):
        x_min = min_value+i*binsize
        x_max = min_value+(i+1)*binsize
        #x_val = 0.5*(x_min+x_max)
        y_val = 0.0
        for ts in values:
            if ts > x_min:
                y_val += 1.0
        hist_x.append(x_min)
        hist_y.append(y_val)
        hist_x.append(x_max)
        hist_y.append(y_val)
        if x_min > 0.0:
            value = norm*erf(math.sqrt(0.5*x_min)) + norm
            hist_expect.append(len(values)-value)
            hist_expect.append(len(values)-value)
        else:
            hist_expect.append(norm)
            hist_expect.append(norm)

    # Show expected distribution (half)
    plt.semilogy(hist_x, hist_y,      'r-', linewidth=2, label='simulations')
    if expect == 'show':
        plt.semilogy(hist_x, hist_expect, 'k-', linewidth=2, label='expectation')
        plt.semilogy(hist_x, hist_y,      'r-', linewidth=2)

    # Set plot
    plt.xlabel('TS')
    plt.ylabel('Number of trials > TS')
    plt.title(title)
    plt.grid(True)
    plt.legend(loc='upper right')

    # Show figure
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ======================================== #
# Display probability density distribution #
# ======================================== #
def dist_pdf(values, nbins, title, plotfile, expect=''):
    """
    Display probability density TS distribution and overplots expectation

    Parameters
    ----------
    values : float
        TS values
    nbins : int
        Number of bins
    title : str
        Plot title
    plotfile : str
        Plot file name
    expect : str, optional
        If 'show', show expectation
    """
    # Set range and adapt number of bins. We make sure that we have a
    # histogram bin centred on 0 that will capture the TS=0 part.
    max_value = max(values)
    binsize   = max_value/nbins
    min_value = 0.5*binsize
    while min_value > min(values):
        min_value -= binsize
        nbins     += 1

    # Create histogram
    _, bins, _ = plt.hist(values, nbins, range=[min_value, max_value],
                          align='mid', facecolor='green',
                          log=True, label='simulations')

    # Create expected distribution (only for positive TS). We compute the
    # full and half of the distribution, as for positively constrained
    # quantities half of the values have TS=0
    x      = []
    y_full = []
    y_half = []
    width  = bins[1]-bins[0]
    norm   = len(values)/(math.sqrt(2.0)*math.sqrt(math.pi))*width
    x.append(0.0)
    y_full.append(len(values))
    y_half.append(0.5*len(values))
    for i in bins:
        ts = i + 0.5*width
        if ts > 0:
            y  = norm*math.pow(ts,-0.5)*math.exp(-0.5*ts)
            x.append(ts)
            y_full.append(y)
            y_half.append(0.5*y)

    # Show expected distribution (half)
    if expect == 'show':
        plt.semilogy(x, y_half, 'ro', linewidth=2, label='expectation')

    # Set plot
    plt.xlabel('TS')
    plt.ylabel('Number of trials')
    plt.title(title)
    plt.grid(True)
    plt.legend(loc='upper right')

    # Show figure
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ==================== #
# Show TS distribution #
# ==================== #
def show_ts_distribution():
    """
    Show TS distribution
    """
    # Set usage string
    usage = ('show_ts_distribution.py [-n bins] [-c column] [-title title] '
             '[-e expect] [-t type] [-p plotfile] file')

    # Set default options
    options = [{'option': '-n', 'value': 30},
               {'option': '-c', 'value': 'TS'},
               {'option': '-title', 'value': 'TS distribution'},
               {'option': '-e', 'value': ''},
               {'option': '-t', 'value': 'pdf'},
               {'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    nbins    = int(options[0]['value'])
    tsname   = options[1]['value']
    title    = options[2]['value']
    expect   = options[3]['value']
    disttype = options[4]['value']
    plotfile = options[5]['value']

    # Read values from CSV file
    values = read_ts(args[0], tsname=tsname)
    print('%d values read.' % len(values))

    # Show histogram
    if disttype == 'pdf':
        dist_pdf(values, nbins, title, plotfile, expect=expect)
    else:
        dist_cdf(values, nbins, title, plotfile, expect=expect)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show TS distribution
    show_ts_distribution()



#! /usr/bin/env python
# ==========================================================================
# Executes analysis workflow
#
# Copyright (C) 2016-2017 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools
import cscripts


# ================ #
# csworkflow class #
# ================ #
class csworkflow(ctools.cscript):
    """
    Executes an analysis workflow.

    The ``csworkflow`` script executes an analysis workflow defined in an
    XML file.
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor.

        Parameters
        ----------
        argv : list of str
            List of IRAF command line parameter strings of the form
            ``parameter=3``.

        Raises
        ------
        TypeError
            An invalid number of command line arguments was provided.
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Set members
        self._workflow = gammalib.GXml()
        self._actors   = []

        # Return
        return


    # Private methods
    #
    # Query all user parameters
    def _get_parameters(self):

        # Load workflow XML file
        xmlfile        = self['inflow'].filename()
        self._workflow = gammalib.GXml(xmlfile.url())

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return

    # Parse the workflow XML file
    def _parse_workflow(self):

        # Get workflow
        workflow = self._workflow.element('workflow')
        
        # Get number of actors
        num_actors = workflow.elements('actor')

        # Initialise actors
        self._actors = []

        # Loop over all actors
        for i in range(num_actors):
        
            # Get actor
            actor = workflow.element('actor', i)

            # Initialise parameters and input actors
            input_parameters  = []
            output_parameters = []
            input_actors      = []
            output_actors     = []

            # Get actor attributes
            name = actor.attribute('name')
            tool = actor.attribute('tool')

            # Get actor input parameters
            if actor.elements('input') > 0:
                actor_inputs = actor.element('input')
                num_inputs   = actor_inputs.elements('parameter')
                for k in range(num_inputs):
                    input_par   = actor_inputs.element('parameter', k)
                    input_name  = input_par.attribute('name')
                    input_value = input_par.attribute('value')
                    input_actor = input_par.attribute('actor')
                    parameter   = {'name': input_name, \
                                   'value': input_value, \
                                   'actor': input_actor}
                    input_parameters.append(parameter)
                    if input_actor != '':
                        if input_actor not in input_actors:
                            input_actors.append(input_actor)

            # Get actor output parameters
            if actor.elements('output') > 0:
                actor_output = actor.element('output')
                num_outputs  = actor_output.elements('parameter')
                for k in range(num_outputs):
                    output_par   = actor_output.element('parameter', k)
                    output_name  = output_par.attribute('name')
                    output_value = output_par.attribute('value')
                    output_actor = output_par.attribute('actor')
                    parameter    = {'name': output_name, \
                                    'value': output_value, \
                                    'actor': output_actor}
                    output_parameters.append(parameter)
                    if output_actor != '':
                        if output_actor not in output_actors:
                            output_actors.append(output_actor)

            # Determine number of dependencies
            num_inputs = len(input_actors)

            # Set actor status
            if num_inputs > 0:
                status = 'waiting for input'
            else:
                status = 'ready'

            # Create actor entry
            entry = {'name': name,
                     'tool': tool,
                     'input_parameters': input_parameters,
                     'input_actors': input_actors,
                     'output_parameters': output_parameters,
                     'output_actors': output_actors,
                     'status': status}

            # Append entry
            self._actors.append(entry)

            # Log information about actors
            self._log_value(gammalib.NORMAL, 'Actor "%s"' % name, tool)

            # Compute list of predecessors
            if num_inputs == 0:
                predecessors = 'none'
            else:
                predecessors = ''
                for k in range(num_inputs):
                    if k > 0:
                        predecessors += ', '
                    predecessors += '"'+input_actors[k]+'"'
 
            # Log predecessors
            self._log_value(gammalib.NORMAL,
                            gammalib.number('  Predecessor', num_inputs),
                            predecessors)

        # Return
        return

    # Execute a workflow
    def _execute_workflow(self):

        # Continue while there are actors
        while len(self._actors) > 0:

            # Find actors which are ready
            num_executed = 0
            for actor in self._actors:

                # If actor is ready then execute it
                if actor['status'] == 'ready':

                    # Log execution start
                    self._log_value(gammalib.NORMAL, 'Execute actor',
                                    '"%s"' % actor['name'])

                    # Execute actor
                    self._execute_actor(actor)

                    # Log execution finish
                    self._log_value(gammalib.NORMAL, 'Finished actor execution',
                                    '"%s"' % actor['name'])

                    # Set actor status to finished
                    actor['status'] = 'finished'

                    # Increment number of executed actors
                    num_executed += 1

            # Break if no actors have been executed
            if num_executed < 1:
                break
            
            # Update actor status
            for actor in self._actors:
                if actor['status'] != 'finished':
                    input_actors = actor['input_actors']
                    ready        = True
                    for input_actor in input_actors:
                        a = self._get_actor(input_actor)
                        if a['status'] != 'finished':
                            ready = False
                            break
                    if ready:
                        actor['status'] = 'ready'

        # Return
        return

    # Execute an actor
    def _execute_actor(self, actor):

        # Log input parameters
        pars = actor['input_parameters']
        for par in pars:
            self._log_value(gammalib.NORMAL, '  Input parameter', '%s=%s' %
                            (par['name'], self._get_parameter_value(par)))

        # Log output parameters
        pars = actor['output_parameters']
        for par in pars:
            self._log_value(gammalib.NORMAL, '  Output parameter', '%s=%s' %
                            (par['name'], self._get_parameter_value(par)))

        # Set actor tool
        if 'tool' in actor:
            tool = actor['tool']
            if tool[:2] == 'ct':
                tool_eval = 'ctools.'+tool+'()'
            elif tool[:2] == 'cs':
                tool_eval = 'cscripts.'+tool+'()'
            else:
                tool_eval = ''
        if tool_eval == '':
            return

        # Create actor tool object
        tool = eval(tool_eval)

        # Set actor input parameters
        pars = actor['input_parameters']
        for par in pars:
            app_par = tool[par['name']]
            app_par.value(self._get_parameter_value(par))
            tool[par['name']] = app_par

        # Set actor output parameters
        pars = actor['output_parameters']
        for par in pars:
            app_par = tool[par['name']]
            app_par.value(self._get_parameter_value(par))
            tool[par['name']] = app_par

        # Execute actor
        tool.logFileOpen()
        tool.execute()
        
        # Return
        return

    # Get the parameter value from the respective actor
    def _get_parameter_value(self, par):

        # If the actor is empty then simply return the value
        if (par['actor'] == ''):
            value = par['value']

        # ... otherwise get the value from the actor
        else:
            found = False
            actor = self._get_actor(par['actor'])
            for output in actor['output_parameters']:
                if output['name'] == par['value']:
                    value = output['value']
                    found = True
                    break
            if not found:
                msg = 'Parameter "'+par['value']+'" is not an output '+ \
                      'parameter of actor "'+par['actor']+'".'
                raise RuntimeError(msg)
        
        # Return value
        return value

    # Get the actor with the specified name
    def _get_actor(self, name):

        # Return actor with specified name
        for actor in self._actors:
            if actor['name'] == name:
                return actor

        # Throw an exception
        msg = 'Actor "'+name+'" not found.'
        raise RuntimeError(msg)
 
        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Parse XML file
        if self._logTerse():
            self._log('\n')
            self._log.header1('Parse workflow XML file')
        self._parse_workflow()

        # Execute workflow
        if self._logTerse():
            self._log('\n')
            self._log.header1('Execute workflow')
        self._execute_workflow()

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csworkflow(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the ctexpcube tool.
#
# Copyright (C) 2014-2020 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import os
import gammalib
import ctools
from testing import test


# ============================= #
# Test class for ctexpcube tool #
# ============================= #
class Test(test):
    """
    Test class for ctexpcube tool
    """
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('ctexpcube')

        # Append tests
        self.append(self._test_cmd, 'Test ctexpcube on command line')
        self.append(self._test_python, 'Test ctexpcube from Python')

        # Return
        return

    # Test ctexpcube on command line
    def _test_cmd(self):
        """
        Test ctexpcube on the command line
        """
        # Set tool name
        ctexpcube = self._tool('ctexpcube')

        # Setup ctexpcube command
        cmd = ctexpcube+' inobs="'+self._events+'"'+ \
                        ' incube="NONE"'+ \
                        ' outcube="ctexpcube_cmd1.fits"'+ \
                        ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                        ' ebinalg="LOG" emin=1.0 emax=100.0 enumbins=5'+ \
                        ' nxpix=20 nypix=20 binsz=0.2'+ \
                        ' coordsys="CEL" proj="CAR" xref=83.63 yref=22.01'+ \
                        ' logfile="ctexpcube_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('ctexpcube_cmd1.fits')

        # Setup ctexpcube command
        cmd = ctexpcube+' inobs="event_file_that_does_not_exist.fits"'+ \
                        ' incube="NONE"'+ \
                        ' outcube="ctexpcube_cmd2.fits"'+ \
                        ' caldb="'+self._caldb+'" irf="'+self._irf+'"'+ \
                        ' ebinalg="LOG" emin=1.0 emax=100.0 enumbins=5'+ \
                        ' nxpix=20 nypix=20 binsz=0.2'+ \
                        ' coordsys="CEL" proj="CAR" xref=83.63 yref=22.01'+ \
                        ' logfile="ctexpcube_cmd2.log" debug=yes chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check ctexpcube --help
        self._check_help(ctexpcube)

        # Return
        return

    # Test ctexpcube from Python
    def _test_python(self):
        """
        Test ctexpcube from Python
        """
        # Allocate ctexpcube
        expcube = ctools.ctexpcube()

        # Check that empty ctexpcube tool holds an exposure cube that has
        # no energy bins
        self._check_cube(expcube.expcube(), nenergies=0)

        # Check that saving does not nothing
        expcube['outcube'] = 'ctexpcube_py0.fits'
        expcube['logfile'] = 'ctexpcube_py0.log'
        expcube.logFileOpen()
        expcube.save()
        self.test_assert(not os.path.isfile('ctexpcube_py0.fits'),
             'Check that no exposure cube has been created')

        # Check that clearing does not lead to an exception or segfault
        expcube.clear()

        # Now set ctexpcube parameters
        expcube['inobs']    = self._events
        expcube['incube']   = 'NONE'
        expcube['caldb']    = self._caldb
        expcube['irf']      = self._irf
        expcube['ebinalg']  = 'LOG'
        expcube['emin']     = 1.0
        expcube['emax']     = 100.0
        expcube['enumbins'] = 5
        expcube['nxpix']    = 20
        expcube['nypix']    = 20
        expcube['binsz']    = 0.2
        expcube['coordsys'] = 'CEL'
        expcube['proj']     = 'CAR'
        expcube['xref']     = 83.63
        expcube['yref']     = 22.01
        expcube['logfile']  = 'ctexpcube_py1.log'
        expcube['outcube']  = 'ctexpcube_py1.fits'
        expcube['chatter']  = 2

        # Run ctexpcube tool
        expcube.logFileOpen()   # Make sure we get a log file
        expcube.run()
        expcube.save()

        # Check result file
        self._check_result_file('ctexpcube_py1.fits')

        # Copy ctexpcube tool
        cpy_expcube = expcube.copy()

        # Check exposure cube of ctexpcube copy
        self._check_cube(cpy_expcube.expcube())

        # Execute copy of ctexpcube tool again, now with a higher chatter
        # level than before
        cpy_expcube['emin']      = 0.2
        cpy_expcube['emax']      = 150.0
        cpy_expcube['outcube']   = 'ctexpcube_py2.fits'
        cpy_expcube['logfile']   = 'ctexpcube_py2.log'
        cpy_expcube['publish']   = True
        cpy_expcube['addbounds'] = True
        cpy_expcube['chatter']   = 3
        cpy_expcube.logFileOpen()  # Needed to get a new log file
        cpy_expcube.execute()

        # Check result file
        self._check_result_file('ctexpcube_py2.fits', nenergies=8)

        # Now clear copy of ctexpcube tool
        cpy_expcube.clear()

        # Check that the cleared copy has also cleared the exposure cube
        self._check_cube(cpy_expcube.expcube(), nenergies=0)

        # Get mixed observation container
        obs = self._obs_mixed()

        # Set-up ctexpcube from observation container
        expcube = ctools.ctexpcube(obs)
        expcube['incube']    = 'NONE'
        expcube['caldb']     = self._caldb
        expcube['irf']       = self._irf
        expcube['ebinalg']   = 'POW'
        expcube['emin']      = 1.0
        expcube['emax']      = 100.0
        expcube['enumbins']  = 5
        expcube['ebingamma'] = 1.0
        expcube['nxpix']     = 20
        expcube['nypix']     = 20
        expcube['binsz']     = 0.2
        expcube['coordsys']  = 'CEL'
        expcube['proj']      = 'CAR'
        expcube['xref']      = 83.63
        expcube['yref']      = 22.01
        expcube['logfile']   = 'ctexpcube_py3.log'
        expcube['outcube']   = 'ctexpcube_py3.fits'
        expcube['chatter']   = 4

        # Execute ctexpcube tool
        expcube.logFileOpen()  # Needed to get a new log file
        expcube.execute()

        # Check result file
        self._check_result_file('ctexpcube_py3.fits')

        # Set-up ctexpcube from event list and counts cube
        expcube = ctools.ctexpcube()
        expcube['inobs']    = self._events
        expcube['incube']   = self._cntcube
        expcube['caldb']    = self._caldb
        expcube['irf']      = self._irf
        expcube['logfile']  = 'ctexpcube_py4.log'
        expcube['outcube']  = 'ctexpcube_py4.fits'
        expcube['chatter']  = 2

        # Execute ctexpcube tool
        expcube.logFileOpen()  # Needed to get a new log file
        expcube.execute()

        # Check result file
        self._check_result_file('ctexpcube_py4.fits')

        # Return
        return

    # Check result file
    def _check_result_file(self, filename, nenergies=6):
        """
        Check result file

        Parameters
        ----------
        filename : str
            Exposure cube file name
        nenergies : int, optional
            Number of energy bins
        """
        # Open exposure cube
        cube = gammalib.GCTACubeExposure(filename)

        # Check cube
        self._check_cube(cube, nenergies=nenergies)

        # Return
        return

    # Check exposure cube
    def _check_cube(self, cube, nenergies=6):
        """
        Check exposure cube

        Parameters
        ----------
        cube : `~gammalib.GCTACubeExposure`
            Exposure cube
        nenergies : int, optional
            Number of energy bins
        """
        # Check dimensions
        self.test_value(len(cube.energies()), nenergies,
             'Check number of energy maps')

        # Return
        return
